"""
Tests for stringutils utility file.
"""

import builtins
import re
import textwrap

import pytest

import salt.utils.stringutils
from tests.support.mock import patch
from tests.support.unit import LOREM_IPSUM


@pytest.fixture(scope="function")
def unicode():
    UNICODE = "中国語 (繁体)"
    STR = BYTES = UNICODE.encode("utf-8")
    return UNICODE, STR, BYTES


@pytest.fixture(scope="function")
def eggs():
    # This is an example of a unicode string with й constructed using two separate
    # code points. Do not modify it.
    eggs = "\u044f\u0438\u0306\u0446\u0430"
    return eggs


@pytest.fixture(scope="function")
def latin1_unicode():
    LATIN1_UNICODE = "räksmörgås"
    LATIN1_BYTES = LATIN1_UNICODE.encode("latin-1")
    return LATIN1_UNICODE, LATIN1_BYTES


@pytest.fixture()
def single_txt():
    return """
# set variable identifying the chroot you work in (used in the prompt below)
if [ -z '$debian_chroot' ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi
"""


@pytest.fixture()
def double_txt():
    return """
# set variable identifying the chroot you work in (used in the prompt below)
if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi
"""


@pytest.fixture()
def single_double_txt():
    return """
# set variable identifying the chroot you work in (used in the prompt below)
if [ -z '$debian_chroot' ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi

# set variable identifying the chroot you work in (used in the prompt below)
if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi
"""


@pytest.fixture()
def single_double_same_line_txt():
    return """
# set variable identifying the chroot you work in (used in the prompt below)
if [ -z '$debian_chroot' ] && [ -r "/etc/debian_chroot" ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi
"""


@pytest.fixture()
def match():
    return """
# set variable identifying the chroot you work in (used in the prompt below)
if [ -z '$debian_chroot' ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi


# set variable identifying the chroot you work in (used in the prompt below)
if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi


# set variable identifying the chroot you work in (used in the prompt below)
if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi


# set variable identifying the chroot you work in (used in the prompt below)
if [ -z '$debian_chroot' ] && [ -r /etc/debian_chroot ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi


# set variable identifying the chroot you work in (used in the prompt below)
if [ -z '$debian_chroot' ] && [ -r "/etc/debian_chroot" ]; then
    debian_chroot=$(cat /etc/debian_chroot)
fi
"""


def test_single_quotes(single_txt, match):
    regex = salt.utils.stringutils.build_whitespace_split_regex(single_txt)
    assert re.search(regex, match)


def test_double_quotes(double_txt, match):
    regex = salt.utils.stringutils.build_whitespace_split_regex(double_txt)
    assert re.search(regex, match)


def test_single_and_double_quotes(single_double_txt, match):
    regex = salt.utils.stringutils.build_whitespace_split_regex(single_double_txt)
    assert re.search(regex, match)


def test_issue_2227(single_double_same_line_txt, match):
    regex = salt.utils.stringutils.build_whitespace_split_regex(
        single_double_same_line_txt
    )
    assert re.search(regex, match)


def test_contains_whitespace():
    does_contain_whitespace = "A brown fox jumped over the red hen."
    does_not_contain_whitespace = "Abrownfoxjumpedovertheredhen."

    assert salt.utils.stringutils.contains_whitespace(does_contain_whitespace) is True
    assert (
        salt.utils.stringutils.contains_whitespace(does_not_contain_whitespace) is False
    )


def test_to_num():
    assert salt.utils.stringutils.to_num("7") == 7
    assert isinstance(salt.utils.stringutils.to_num("7"), int)
    assert salt.utils.stringutils.to_num("7.0") == 7
    assert isinstance(salt.utils.stringutils.to_num("7.0"), float)
    assert salt.utils.stringutils.to_num("Seven") == "Seven"
    assert isinstance(salt.utils.stringutils.to_num("Seven"), str)


def test_to_none():
    assert salt.utils.stringutils.to_none("") is None
    assert salt.utils.stringutils.to_none("  ") is None
    # Ensure that we do not inadvertently convert certain strings or 0 to None
    assert salt.utils.stringutils.to_none("None") is not None
    assert salt.utils.stringutils.to_none(0) is not None


def test_is_binary():
    assert salt.utils.stringutils.is_binary(LOREM_IPSUM) is False
    # Also test bytestring
    assert (
        salt.utils.stringutils.is_binary(salt.utils.stringutils.is_binary(LOREM_IPSUM))
        is False
    )

    zero_str = "{}{}".format(LOREM_IPSUM, "\0")
    assert salt.utils.stringutils.is_binary(zero_str) is True
    # Also test bytestring
    assert (
        salt.utils.stringutils.is_binary(salt.utils.stringutils.to_bytes(zero_str))
        is True
    )

    # To to ensure safe exit if str passed doesn't evaluate to True
    assert salt.utils.stringutils.is_binary("") is False
    assert salt.utils.stringutils.is_binary(b"") is False

    nontext = 3 * "".join([chr(x) for x in range(1, 32) if x not in (8, 9, 10, 12, 13)])
    almost_bin_str = f"{LOREM_IPSUM[:100]}{nontext[:42]}"

    assert salt.utils.stringutils.is_binary(almost_bin_str) is False
    # Also test bytestring
    assert (
        salt.utils.stringutils.is_binary(
            salt.utils.stringutils.to_bytes(almost_bin_str)
        )
        is False
    )

    bin_str = almost_bin_str + "\x01"
    assert salt.utils.stringutils.is_binary(bin_str) is True
    # Also test bytestring
    assert (
        salt.utils.stringutils.is_binary(salt.utils.stringutils.to_bytes(bin_str))
        is True
    )


def test_to_str(unicode):
    UNICODE = unicode[0]
    BYTES = unicode[2]

    for x in (123, (1, 2, 3), [1, 2, 3], {1: 23}, None):
        assert pytest.raises(TypeError, salt.utils.stringutils.to_str, x)
    assert salt.utils.stringutils.to_str("plugh") == "plugh"
    assert salt.utils.stringutils.to_str("áéíóúý", "utf-8") == "áéíóúý"
    assert salt.utils.stringutils.to_str(BYTES, "utf-8") == UNICODE
    assert salt.utils.stringutils.to_str(bytearray(BYTES), "utf-8") == UNICODE
    # Test situation when a minion returns incorrect utf-8 string because of... million reasons
    ut2 = b"\x9c"
    assert pytest.raises(
        UnicodeDecodeError, salt.utils.stringutils.to_str, ut2, "utf-8"
    )
    assert salt.utils.stringutils.to_str(ut2, "utf-8", "replace") == "\ufffd"
    assert pytest.raises(
        UnicodeDecodeError,
        salt.utils.stringutils.to_str,
        bytearray(ut2),
        "utf-8",
    )
    assert salt.utils.stringutils.to_str(bytearray(ut2), "utf-8", "replace") == "\ufffd"


def test_to_bytes(unicode):
    UNICODE = unicode[0]
    BYTES = unicode[2]

    for x in (123, (1, 2, 3), [1, 2, 3], {1: 23}, None):
        assert pytest.raises(TypeError, salt.utils.stringutils.to_bytes, x)

    assert salt.utils.stringutils.to_bytes("xyzzy") == b"xyzzy"
    assert salt.utils.stringutils.to_bytes(BYTES) == BYTES
    assert salt.utils.stringutils.to_bytes(bytearray(BYTES)) == BYTES
    assert salt.utils.stringutils.to_bytes(UNICODE, "utf-8") == BYTES

    # Test utf-8 fallback with ascii default encoding
    with patch.object(builtins, "__salt_system_encoding__", "ascii"):
        assert salt.utils.stringutils.to_bytes("Ψ") == b"\xce\xa8"


def test_to_unicode(eggs, unicode, latin1_unicode):
    UNICODE = unicode[0]
    BYTES = unicode[2]
    LATIN1_UNICODE = latin1_unicode[0]
    LATIN1_BYTES = latin1_unicode[1]

    assert salt.utils.stringutils.to_unicode(eggs, normalize=True) == "яйца"
    assert salt.utils.stringutils.to_unicode(eggs, normalize=False) != "яйца"
    assert (
        salt.utils.stringutils.to_unicode(LATIN1_BYTES, encoding="latin-1")
        == LATIN1_UNICODE
    )
    assert salt.utils.stringutils.to_unicode("plugh") == "plugh"
    assert salt.utils.stringutils.to_unicode("áéíóúý") == "áéíóúý"
    assert salt.utils.stringutils.to_unicode(BYTES, "utf-8") == UNICODE
    assert salt.utils.stringutils.to_unicode(bytearray(BYTES), "utf-8") == UNICODE


def test_to_unicode_multi_encoding(latin1_unicode):
    LATIN1_UNICODE = latin1_unicode[0]
    LATIN1_BYTES = latin1_unicode[1]
    assert (
        salt.utils.stringutils.to_unicode(LATIN1_BYTES, encoding=("utf-8", "latin1"))
        == LATIN1_UNICODE
    )


def test_build_whitespace_split_regex():
    # With 3.7+,  re.escape only escapes special characters, no longer
    # escaping all characters other than ASCII letters, numbers and
    # underscores.  This includes commas.
    expected_regex = (
        "(?m)^(?:[\\s]+)?Lorem(?:[\\s]+)?ipsum(?:[\\s]+)?dolor(?:[\\s]+)?sit(?:[\\s]+)?amet,"
        "(?:[\\s]+)?$"
    )
    assert (
        salt.utils.stringutils.build_whitespace_split_regex(
            " ".join(LOREM_IPSUM.split()[:5])
        )
        == expected_regex
    )


def test_get_context():
    expected_context = textwrap.dedent(
        """\
        ---
        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Quisque eget urna a arcu lacinia sagittis.
        Sed scelerisque, lacus eget malesuada vestibulum, justo diam facilisis tortor, in sodales dolor
        [...]
        ---"""
    )
    assert (
        salt.utils.stringutils.get_context(LOREM_IPSUM, 1, num_lines=1)
        == expected_context
    )


def test_get_context_has_enough_context():
    template = "1\n2\n3\n4\n5\n6\n7\n8\n9\na\nb\nc\nd\ne\nf"
    context = salt.utils.stringutils.get_context(template, 8)
    expected = "---\n[...]\n3\n4\n5\n6\n7\n8\n9\na\nb\nc\nd\n[...]\n---"
    assert expected == context


def test_get_context_at_top_of_file():
    template = "1\n2\n3\n4\n5\n6\n7\n8\n9\na\nb\nc\nd\ne\nf"
    context = salt.utils.stringutils.get_context(template, 1)
    expected = "---\n1\n2\n3\n4\n5\n6\n[...]\n---"
    assert expected == context


def test_get_context_at_bottom_of_file():
    template = "1\n2\n3\n4\n5\n6\n7\n8\n9\na\nb\nc\nd\ne\nf"
    context = salt.utils.stringutils.get_context(template, 15)
    expected = "---\n[...]\na\nb\nc\nd\ne\nf\n---"
    assert expected == context


def test_get_context_2_context_lines():
    template = "1\n2\n3\n4\n5\n6\n7\n8\n9\na\nb\nc\nd\ne\nf"
    context = salt.utils.stringutils.get_context(template, 8, num_lines=2)
    expected = "---\n[...]\n6\n7\n8\n9\na\n[...]\n---"
    assert expected == context


def test_get_context_with_marker():
    template = "1\n2\n3\n4\n5\n6\n7\n8\n9\na\nb\nc\nd\ne\nf"
    context = salt.utils.stringutils.get_context(
        template, 8, num_lines=2, marker=" <---"
    )
    expected = "---\n[...]\n6\n7\n8 <---\n9\na\n[...]\n---"
    assert expected == context


def test_expr_match():
    val = "foo/bar/baz"
    # Exact match
    assert salt.utils.stringutils.expr_match(val, val) is True
    # Glob match
    assert salt.utils.stringutils.expr_match(val, "foo/*/baz") is True
    # Glob non-match
    assert salt.utils.stringutils.expr_match(val, "foo/*/bar") is False
    # Regex match
    assert salt.utils.stringutils.expr_match(val, r"foo/\w+/baz") is True
    # Regex non-match
    assert salt.utils.stringutils.expr_match(val, r"foo/\w/baz") is False


def test_check_whitelist_blacklist():
    """
    Ensure that whitelist matching works on both PY2 and PY3
    """
    whitelist = ["one/two/three", r"web[0-9]"]
    blacklist = ["four/five/six", r"web[5-9]"]

    # Tests with string whitelist/blacklist
    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_one",
            whitelist=whitelist[1],
            blacklist=None,
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_one",
            whitelist=whitelist[1],
            blacklist=[],
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web1",
            whitelist=whitelist[1],
            blacklist=None,
        )
        is True
    )
    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web1",
            whitelist=whitelist[1],
            blacklist=[],
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=None,
            blacklist=blacklist[1],
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=[],
            blacklist=blacklist[1],
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_five",
            whitelist=None,
            blacklist=blacklist[1],
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_five",
            whitelist=[],
            blacklist=blacklist[1],
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=whitelist[1],
            blacklist=blacklist[1],
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web4",
            whitelist=whitelist[1],
            blacklist=blacklist[1],
        )
        is True
    )

    # Tests with list whitelist/blacklist
    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_one",
            whitelist=whitelist,
            blacklist=None,
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_one",
            whitelist=whitelist,
            blacklist=[],
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web1",
            whitelist=whitelist,
            blacklist=None,
        )
        is True
    )

    assert salt.utils.stringutils.check_whitelist_blacklist(
        "web1",
        whitelist=whitelist,
        blacklist=[],
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=None,
            blacklist=blacklist,
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=[],
            blacklist=blacklist,
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_five",
            whitelist=None,
            blacklist=blacklist,
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_five",
            whitelist=[],
            blacklist=blacklist,
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=whitelist,
            blacklist=blacklist,
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web4",
            whitelist=whitelist,
            blacklist=blacklist,
        )
        is True
    )

    # Tests with set whitelist/blacklist
    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_one",
            whitelist=set(whitelist),
            blacklist=None,
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_one",
            whitelist=set(whitelist),
            blacklist=set(),
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web1",
            whitelist=set(whitelist),
            blacklist=None,
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web1",
            whitelist=set(whitelist),
            blacklist=set(),
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=None,
            blacklist=set(blacklist),
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=set(),
            blacklist=set(blacklist),
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_five",
            whitelist=None,
            blacklist=set(blacklist),
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web_five",
            whitelist=set(),
            blacklist=set(blacklist),
        )
        is True
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web5",
            whitelist=set(whitelist),
            blacklist=set(blacklist),
        )
        is False
    )

    assert (
        salt.utils.stringutils.check_whitelist_blacklist(
            "web4",
            whitelist=set(whitelist),
            blacklist=set(blacklist),
        )
        is True
    )

    # Test with invalid type for whitelist/blacklist
    assert pytest.raises(
        TypeError,
        salt.utils.stringutils.check_whitelist_blacklist,
        "foo",
        whitelist=123,
    )
    assert pytest.raises(
        TypeError,
        salt.utils.stringutils.check_whitelist_blacklist,
        "foo",
        blacklist=123,
    )


def test_check_include_exclude_empty():
    assert salt.utils.stringutils.check_include_exclude("/some/test") is True


def test_check_include_exclude_exclude():
    assert (
        salt.utils.stringutils.check_include_exclude("/some/test", None, "*test*")
        is False
    )


def test_check_include_exclude_exclude_list():
    assert (
        salt.utils.stringutils.check_include_exclude("/some/test", None, ["*test"])
        is False
    )


def test_check_include_exclude_exclude_include():
    assert (
        salt.utils.stringutils.check_include_exclude("/some/test", "*test*", "/some/")
        is True
    )


def test_check_include_exclude_regex():
    assert (
        salt.utils.stringutils.check_include_exclude(
            "/some/test", None, "E@/some/(test|other)"
        )
        is False
    )


@pytest.mark.parametrize(
    "unit",
    [
        "B",
        "K",
        "KB",
        "KiB",
        "M",
        "MB",
        "MiB",
        "G",
        "GB",
        "GiB",
        "T",
        "TB",
        "TiB",
        "P",
        "PB",
        "PiB",
        "E",
        "EB",
        "EiB",
        "Z",
        "ZB",
        "ZiB",
        "Y",
        "YB",
        "YiB",
    ],
)
def test_human_to_bytes(unit):
    # first multiplier is IEC/binary
    # second multiplier is metric/decimal
    conversion = {
        "B": (1, 1),
        "K": (2**10, 10**3),
        "M": (2**20, 10**6),
        "G": (2**30, 10**9),
        "T": (2**40, 10**12),
        "P": (2**50, 10**15),
        "E": (2**60, 10**18),
        "Z": (2**70, 10**21),
        "Y": (2**80, 10**24),
    }

    idx = 0
    if len(unit) == 2:
        idx = 1

    # pull out the multipliers for the units
    multiplier = conversion[unit.upper()[0]][idx]
    iec = conversion[unit.upper()[0]][0]

    vals = [32]
    # don't calculate a half a byte
    if unit != "B":
        # otherwise, test a float as well
        vals.append(64.5)

    for val in vals:
        # calculate KB, MB, GB, etc. as 1024 instead of 1000 (legacy use)
        assert salt.utils.stringutils.human_to_bytes(f"{val}{unit}") == val * iec
        assert salt.utils.stringutils.human_to_bytes(f"{val} {unit}") == val * iec
        # handle metric (KB, MB, GB, etc.) per standard
        assert (
            salt.utils.stringutils.human_to_bytes(f"{val}{unit}", handle_metric=True)
            == val * multiplier
        )
        assert (
            salt.utils.stringutils.human_to_bytes(f"{val} {unit}", handle_metric=True)
            == val * multiplier
        )


def test_human_to_bytes_edge_cases():
    # no unit - bytes
    assert salt.utils.stringutils.human_to_bytes("32") == 32
    # no unit - default MB
    assert salt.utils.stringutils.human_to_bytes("32", default_unit="M") == 32 * 2**20
    # bad value
    assert salt.utils.stringutils.human_to_bytes("32-1") == 0
    assert salt.utils.stringutils.human_to_bytes("3.4.MB") == 0
    assert salt.utils.stringutils.human_to_bytes("") == 0
    assert salt.utils.stringutils.human_to_bytes("bytes") == 0
    # bad unit
    assert salt.utils.stringutils.human_to_bytes("32gigajammers") == 0
    assert salt.utils.stringutils.human_to_bytes("512bytes") == 0
    assert salt.utils.stringutils.human_to_bytes("4 Kbytes") == 0
    assert salt.utils.stringutils.human_to_bytes("9ib") == 0
    assert salt.utils.stringutils.human_to_bytes("2HB") == 0



"""
Virtual machine image management tools
"""

import logging
import os
import shutil
import tempfile
import uuid

import salt.config
import salt.crypt
import salt.syspaths
import salt.utils.cloud
import salt.utils.files

# Set up logging
log = logging.getLogger(__name__)

# Don't shadow built-in's.
__func_alias__ = {"apply_": "apply"}


def _file_or_content(file_):
    if os.path.exists(file_):
        with salt.utils.files.fopen(file_) as fic:
            return salt.utils.stringutils.to_unicode(fic.read())
    return file_


def prep_bootstrap(mpt):
    """
    Update and get the random script to a random place

    CLI Example:

    .. code-block:: bash

        salt '*' seed.prep_bootstrap /tmp

    """
    # Verify that the boostrap script is downloaded
    bs_ = __salt__["config.gather_bootstrap_script"]()
    fpd_ = os.path.join(mpt, "tmp", f"{uuid.uuid4()}")
    if not os.path.exists(fpd_):
        os.makedirs(fpd_)
    os.chmod(fpd_, 0o700)
    fp_ = os.path.join(fpd_, os.path.basename(bs_))
    # Copy script into tmp
    shutil.copy(bs_, fp_)
    tmppath = fpd_.replace(mpt, "")
    return fp_, tmppath


def _mount(path, ftype, root=None):
    mpt = None
    if ftype == "block":
        mpt = tempfile.mkdtemp()
        if not __salt__["mount.mount"](mpt, path):
            os.rmdir(mpt)
            return None
    elif ftype == "dir":
        return path
    elif ftype == "file":
        if "guestfs.mount" in __salt__:
            util = "guestfs"
        elif "qemu_nbd.init" in __salt__:
            util = "qemu_nbd"
        else:
            return None
        mpt = __salt__["mount.mount"](path, device=root, util=util)
        if not mpt:
            return None
    return mpt


def _umount(mpt, disk, ftype):
    if ftype == "block":
        __salt__["mount.umount"](mpt)
        os.rmdir(mpt)
    elif ftype == "file":
        util = None
        if "guestfs.mount" in __salt__:
            util = "guestfs"
        elif "qemu_nbd.init" in __salt__:
            util = "qemu_nbd"
        if util:
            __salt__["mount.umount"](mpt, device=disk, util=util)


def apply_(
    path,
    id_=None,
    config=None,
    approve_key=True,
    install=True,
    prep_install=False,
    pub_key=None,
    priv_key=None,
    mount_point=None,
):
    """
    Seed a location (disk image, directory, or block device) with the
    minion config, approve the minion's key, and/or install salt-minion.

    CLI Example:

    .. code-block:: bash

        salt 'minion' seed.apply path id [config=config_data] \\
                [gen_key=(true|false)] [approve_key=(true|false)] \\
                [install=(true|false)]

    path
        Full path to the directory, device, or disk image  on the target
        minion's file system.

    id
        Minion id with which to seed the path.

    config
        Minion configuration options. By default, the 'master' option is set to
        the target host's 'master'.

    approve_key
        Request a pre-approval of the generated minion key. Requires
        that the salt-master be configured to either auto-accept all keys or
        expect a signing request from the target host. Default: true.

    install
        Install salt-minion, if absent. Default: true.

    prep_install
        Prepare the bootstrap script, but don't run it. Default: false
    """
    stats = __salt__["file.stats"](path, follow_symlinks=True)
    if not stats:
        return f"{path} does not exist"
    ftype = stats["type"]
    path = stats["target"]
    log.debug("Mounting %s at %s", ftype, path)
    try:
        os.makedirs(path)
    except OSError:
        # The directory already exists
        pass

    mpt = _mount(path, ftype, mount_point)

    if not mpt:
        return f"{path} could not be mounted"

    tmp = os.path.join(mpt, "tmp")
    log.debug("Attempting to create directory %s", tmp)
    try:
        os.makedirs(tmp)
    except OSError:
        if not os.path.isdir(tmp):
            raise
    cfg_files = mkconfig(
        config,
        tmp=tmp,
        id_=id_,
        approve_key=approve_key,
        pub_key=pub_key,
        priv_key=priv_key,
    )

    if _check_install(mpt):
        # salt-minion is already installed, just move the config and keys
        # into place
        log.info("salt-minion pre-installed on image, configuring as %s", id_)
        minion_config = salt.config.minion_config(cfg_files["config"])
        pki_dir = minion_config["pki_dir"]
        if not os.path.isdir(os.path.join(mpt, pki_dir.lstrip("/"))):
            __salt__["file.makedirs"](os.path.join(mpt, pki_dir.lstrip("/"), ""))
        os.rename(
            cfg_files["privkey"], os.path.join(mpt, pki_dir.lstrip("/"), "minion.pem")
        )
        os.rename(
            cfg_files["pubkey"], os.path.join(mpt, pki_dir.lstrip("/"), "minion.pub")
        )
        os.rename(cfg_files["config"], os.path.join(mpt, "etc/salt/minion"))
        res = True
    elif install:
        log.info("Attempting to install salt-minion to %s", mpt)
        res = _install(mpt)
    elif prep_install:
        log.error(
            "The prep_install option is no longer supported. Please use "
            "the bootstrap script installed with Salt, located at %s.",
            salt.syspaths.BOOTSTRAP,
        )
        res = False
    else:
        log.warning("No useful action performed on %s", mpt)
        res = False

    _umount(mpt, path, ftype)
    return res


def mkconfig(
    config=None, tmp=None, id_=None, approve_key=True, pub_key=None, priv_key=None
):
    """
    Generate keys and config and put them in a tmp directory.

    pub_key
        absolute path or file content of an optional preseeded salt key

    priv_key
        absolute path or file content of an optional preseeded salt key

    CLI Example:

    .. code-block:: bash

        salt 'minion' seed.mkconfig [config=config_data] [tmp=tmp_dir] \\
                [id_=minion_id] [approve_key=(true|false)]
    """
    if tmp is None:
        tmp = tempfile.mkdtemp()
    if config is None:
        config = {}
    if "master" not in config and __opts__["master"] != "salt":
        config["master"] = __opts__["master"]
    if id_:
        config["id"] = id_

    # Write the new minion's config to a tmp file
    tmp_config = os.path.join(tmp, "minion")
    with salt.utils.files.fopen(tmp_config, "w+") as fp_:
        fp_.write(salt.utils.cloud.salt_config_to_yaml(config))

    # Generate keys for the minion
    pubkeyfn = os.path.join(tmp, "minion.pub")
    privkeyfn = os.path.join(tmp, "minion.pem")
    preseeded = pub_key and priv_key
    if preseeded:
        log.debug("Writing minion.pub to %s", pubkeyfn)
        log.debug("Writing minion.pem to %s", privkeyfn)
        with salt.utils.files.fopen(pubkeyfn, "w") as fic:
            fic.write(salt.utils.stringutils.to_str(_file_or_content(pub_key)))
        with salt.utils.files.fopen(privkeyfn, "w") as fic:
            fic.write(salt.utils.stringutils.to_str(_file_or_content(priv_key)))
        os.chmod(pubkeyfn, 0o600)
        os.chmod(privkeyfn, 0o600)
    else:
        salt.crypt.gen_keys(tmp, "minion", 2048)
    if approve_key and not preseeded:
        with salt.utils.files.fopen(pubkeyfn) as fp_:
            pubkey = salt.utils.stringutils.to_unicode(fp_.read())
            __salt__["pillar.ext"]({"virtkey": [id_, pubkey]})

    return {"config": tmp_config, "pubkey": pubkeyfn, "privkey": privkeyfn}


def _install(mpt):
    """
    Determine whether salt-minion is installed and, if not,
    install it.
    Return True if install is successful or already installed.
    """
    _check_resolv(mpt)
    boot_, tmppath = prep_bootstrap(mpt) or salt.syspaths.BOOTSTRAP
    # Exec the chroot command
    cmd = "if type salt-minion; then exit 0; "
    cmd += "else sh {} -c /tmp; fi".format(os.path.join(tmppath, "bootstrap-salt.sh"))
    return not __salt__["cmd.run_chroot"](mpt, cmd, python_shell=True)["retcode"]


def _check_resolv(mpt):
    """
    Check that the resolv.conf is present and populated
    """
    resolv = os.path.join(mpt, "etc/resolv.conf")
    replace = False
    if os.path.islink(resolv):
        resolv = os.path.realpath(resolv)
        if not os.path.isdir(os.path.dirname(resolv)):
            os.makedirs(os.path.dirname(resolv))
    if not os.path.isfile(resolv):
        replace = True
    if not replace:
        with salt.utils.files.fopen(resolv, "rb") as fp_:
            conts = salt.utils.stringutils.to_unicode(fp_.read())
            if "nameserver" not in conts:
                replace = True
            if "nameserver 127.0.0.1" in conts:
                replace = True
    if replace:
        shutil.copy("/etc/resolv.conf", resolv)


def _check_install(root):
    sh_ = "/bin/sh"
    if os.path.isfile(os.path.join(root, "bin/bash")):
        sh_ = "/bin/bash"

    cmd = "if ! type salt-minion; then exit 1; fi"
    cmd = f"chroot '{root}' {sh_} -c '{cmd}'"

    return not __salt__["cmd.retcode"](cmd, output_loglevel="quiet", python_shell=True)



import os
import re
import shutil
import subprocess
import time
import winreg

import psutil
import pytest

pytest_plugins = ["helpers_namespace"]

# \n characters are converted to os.linesep
existing_content = [
    "# Existing config from test suite line 1/6\n",
    "master: existing_master\n",
    "# Existing config from test suite line 2/6\n",
    "id: existing_minion\n",
    "# Existing config from test suite line 3/6\n",
    "# Existing config from test suite line 4/6\n",
    "# Existing config from test suite line 5/6\n",
    "# Existing config from test suite line 6/6\n",
]

# \n characters are converted to os.linesep
custom_content = [
    "# Custom config from test suite line 1/6\n",
    "master: custom_master\n",
    "# Custom config from test suite line 2/6\n",
    "id: custom_minion\n",
    "# Custom config from test suite line 3/6\n",
    "# Custom config from test suite line 4/6\n",
    "# Custom config from test suite line 5/6\n",
    "# Custom config from test suite line 6/6\n",
]

# \n characters are converted to os.linesep
old_content = [
    "# Old config from test suite line 1/6\n",
    "master: old_master\n",
    "# Old config from test suite line 2/6\n",
    "id: old_minion\n",
    "# Old config from test suite line 3/6\n",
    "# Old config from test suite line 4/6\n",
    "# Old config from test suite line 5/6\n",
    "# Old config from test suite line 6/6\n",
]

INST_DIR = r"C:\Program Files\Salt Project\Salt"
DATA_DIR = r"C:\ProgramData\Salt Project\Salt"
SYSTEM_DRIVE = os.environ.get("SystemDrive")
OLD_DIR = f"{SYSTEM_DRIVE}\\salt"
SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
INST_BIN = rf"{SCRIPT_DIR}\test-setup.exe"
PROCESSES = [
    os.path.basename(INST_BIN),
    "uninst.exe",
    "Un.exe",
    "Un_A.exe",
    "Un_B.exe",
    "Un_C.exe",
    "Un_D.exe",
    "Un_D.exe",
    "Un_F.exe",
    "Un_G.exe",
]


def reg_key_exists(hive=winreg.HKEY_LOCAL_MACHINE, key=None):
    """
    Helper function to determine if a registry key exists. It does this by
    opening the key. If the connection is successful, the key exists. Otherwise
    an error is returned, which means the key does not exist
    """
    try:
        with winreg.OpenKey(hive, key, 0, winreg.KEY_READ):
            return True
    except:
        return False


def delete_key(hive=winreg.HKEY_LOCAL_MACHINE, key=None):
    if reg_key_exists(hive=hive, key=key):
        parent, _, base = key.rpartition("\\")
        with winreg.OpenKey(hive, parent, 0, winreg.KEY_ALL_ACCESS) as reg:
            winreg.DeleteKey(reg, base)
    assert not reg_key_exists(hive=hive, key=key)


def pytest_configure():
    pytest.DATA_DIR = DATA_DIR
    pytest.INST_DIR = INST_DIR
    pytest.INST_BIN = INST_BIN
    pytest.OLD_DIR = OLD_DIR
    pytest.SCRIPT_DIR = SCRIPT_DIR
    pytest.EXISTING_CONTENT = existing_content
    pytest.CUSTOM_CONTENT = custom_content
    pytest.OLD_CONTENT = old_content


def clean_fragments(inst_dir=INST_DIR):
    # Remove root_dir
    if os.path.exists(DATA_DIR):
        shutil.rmtree(DATA_DIR)
    assert not os.path.exists(DATA_DIR)

    # Remove install dir
    if os.path.exists(inst_dir):
        shutil.rmtree(inst_dir)
    assert not os.path.exists(inst_dir)

    # Remove old salt dir (C:\salt)
    if os.path.exists(OLD_DIR):
        shutil.rmtree(OLD_DIR)
    assert not os.path.exists(OLD_DIR)

    # Remove custom config
    if os.path.exists(rf"{SCRIPT_DIR}\custom_conf"):
        os.remove(rf"{SCRIPT_DIR}\custom_conf")
    assert not os.path.exists(rf"{SCRIPT_DIR}\custom_conf")

    # Remove registry entries
    delete_key(key="SOFTWARE\\Salt Project\\Salt")
    assert not reg_key_exists(
        hive=winreg.HKEY_LOCAL_MACHINE, key="SOFTWARE\\Salt Project\\Salt"
    )

    delete_key(key="SOFTWARE\\Salt Project")
    assert not reg_key_exists(
        hive=winreg.HKEY_LOCAL_MACHINE, key="SOFTWARE\\Salt Project"
    )

    return True


@pytest.helpers.register
def clean_env(inst_dir=INST_DIR, timeout=300):
    # Let's make sure none of the install/uninstall processes are running
    for proc in PROCESSES:
        try:
            assert proc not in (p.name() for p in psutil.process_iter())
        except psutil.NoSuchProcess:
            continue

    # Uninstall existing installation
    # Run the uninstaller.
    for uninst_bin in [f"{inst_dir}\\uninst.exe", f"{OLD_DIR}\\uninst.exe"]:
        if os.path.exists(uninst_bin):
            install_dir = os.path.dirname(uninst_bin)
            cmd = [f'"{uninst_bin}"', "/S", "/delete-root-dir", "/delete-install-dir"]
            run_command(cmd)

            # Uninst.exe launches a 2nd binary (Un.exe or Un_*.exe)
            # Let's get the name of the process
            proc_name = ""
            for proc in PROCESSES:
                try:
                    if proc in (p.name() for p in psutil.process_iter()):
                        proc_name = proc
                except psutil.NoSuchProcess:
                    continue

            # We need to give the process time to exit
            if proc_name:
                elapsed_time = 0
                while elapsed_time < timeout:
                    try:
                        if proc_name not in (p.name() for p in psutil.process_iter()):
                            break
                    except psutil.NoSuchProcess:
                        continue
                    elapsed_time += 0.1
                    time.sleep(0.1)

            assert clean_fragments(inst_dir=install_dir)

    return True


@pytest.helpers.register
def existing_config():
    # Create an existing config
    if not os.path.exists(f"{DATA_DIR}\\conf"):
        os.makedirs(f"{DATA_DIR}\\conf")
    with open(f"{DATA_DIR}\\conf\\minion", "w") as f:
        # \n characters are converted to os.linesep
        f.writelines(existing_content)


@pytest.helpers.register
def custom_config():
    conf_file = rf"{SCRIPT_DIR}\custom_conf"
    if os.path.exists(conf_file):
        os.remove(conf_file)
    # Create a custom config
    with open(conf_file, "w") as f:
        # \n characters are converted to os.linesep
        f.writelines(custom_content)
    assert os.path.exists(conf_file)
    return conf_file


@pytest.helpers.register
def old_install():
    # Create old binaries, don't have to be valid exe's
    if not os.path.exists(f"{OLD_DIR}\\bin"):
        os.makedirs(f"{OLD_DIR}\\bin")
    with open(f"{OLD_DIR}\\bin\\python.exe", "w") as f:
        f.write("binary data")
    with open(f"{OLD_DIR}\\bin\\ssm.exe", "w") as f:
        f.write("binary data")

    # Create an old config
    if not os.path.exists(f"{OLD_DIR}\\conf"):
        os.makedirs(f"{OLD_DIR}\\conf")
    with open(f"{OLD_DIR}\\conf\\minion", "w") as f:
        # \n characters are converted to os.linesep
        f.writelines(old_content)

    assert os.path.exists(f"{OLD_DIR}\\bin\\python.exe")
    assert os.path.exists(f"{OLD_DIR}\\bin\\ssm.exe")
    assert os.path.exists(f"{OLD_DIR}\\conf\\minion")


@pytest.helpers.register
def install_salt(args):
    """
    Cleans the environment and installs salt with passed arguments
    """
    cmd = [f'"{INST_BIN}"']
    if isinstance(args, str):
        cmd.append(args)
    elif isinstance(args, list):
        cmd.extend(args)
    else:
        raise TypeError(f"Invalid args format: {args}")
    run_command(cmd)

    # Let's make sure none of the install/uninstall processes are running
    try:
        assert os.path.basename(INST_BIN) not in (
            p.name() for p in psutil.process_iter()
        )
    except psutil.NoSuchProcess:
        pass


def is_file_locked(path):
    """
    Try to see if a file is locked
    """
    if not (os.path.exists(path)):
        return False
    try:
        f = open(path)
        f.close()
    except OSError:
        return True
    return False


@pytest.helpers.register
def run_command(cmd_args, timeout=300):
    if isinstance(cmd_args, list):
        cmd_args = " ".join(cmd_args)

    bin_file = re.findall(r'"(.*?)"', cmd_args)[0]

    elapsed_time = 0
    while (
        os.path.exists(bin_file) and is_file_locked(bin_file) and elapsed_time < timeout
    ):
        elapsed_time += 0.1
        time.sleep(0.1)

    proc = subprocess.Popen(cmd_args, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

    elapsed_time = 0
    while (
        os.path.exists(bin_file) and is_file_locked(bin_file) and elapsed_time < timeout
    ):
        elapsed_time += 0.1
        time.sleep(0.1)

    try:
        out, err = proc.communicate(timeout=timeout)
        assert proc.returncode == 0
    except subprocess.TimeoutExpired:
        # This hides the hung installer/uninstaller problem
        proc.kill()
        out = "process killed"

    return out



import pytest

try:
    import win32com.client

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False

import salt.utils.win_update as win_update
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.skipif(not HAS_WIN32, reason="Requires Win32 libraries"),
]


def test_available_no_updates():
    """
    Test installed when there are no updates on the system
    """
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(win_update.WindowsUpdateAgent, "refresh", autospec=True):
        wua = win_update.WindowsUpdateAgent(online=False)
        wua._updates = []

        available_updates = wua.available()

        assert available_updates.updates.Add.call_count == 0


def test_available_no_updates_empty_objects():
    """
    Test installed when there are no updates on the system
    """
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(win_update.WindowsUpdateAgent, "refresh", autospec=True):
        wua = win_update.WindowsUpdateAgent(online=False)
        wua._updates = [win32com.client.CDispatch, win32com.client.CDispatch]

        available_updates = wua.available()

        assert available_updates.updates.Add.call_count == 0


def test_installed_no_updates():
    """
    Test installed when there are no updates on the system
    """
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(win_update.WindowsUpdateAgent, "refresh", autospec=True):
        wua = win_update.WindowsUpdateAgent(online=False)
        wua._updates = []

        installed_updates = wua.installed()

        assert installed_updates.updates.Add.call_count == 0


def test_installed_no_updates_installed():
    """
    Test installed when there are no Installed updates on the system
    """
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(win_update.WindowsUpdateAgent, "refresh", autospec=True):
        wua = win_update.WindowsUpdateAgent(online=False)

        wua._updates = [
            MagicMock(IsInstalled=False),
            MagicMock(IsInstalled=False),
            MagicMock(IsInstalled=False),
        ]

        installed_updates = wua.installed()

        assert installed_updates.updates.Add.call_count == 0


def test_installed_updates_all_installed():
    """
    Test installed when all updates on the system are Installed
    """
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(win_update.WindowsUpdateAgent, "refresh", autospec=True):
        wua = win_update.WindowsUpdateAgent(online=False)

        wua._updates = [
            MagicMock(IsInstalled=True),
            MagicMock(IsInstalled=True),
            MagicMock(IsInstalled=True),
        ]

        installed_updates = wua.installed()

        assert installed_updates.updates.Add.call_count == 3


def test_installed_updates_some_installed():
    """
    Test installed when some updates are installed on the system
    """
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(win_update.WindowsUpdateAgent, "refresh", autospec=True):
        wua = win_update.WindowsUpdateAgent(online=False)

        wua._updates = [
            MagicMock(IsInstalled=True),
            MagicMock(IsInstalled=False),
            MagicMock(IsInstalled=True),
            MagicMock(IsInstalled=False),
            MagicMock(IsInstalled=True),
        ]

        installed_updates = wua.installed()

        assert installed_updates.updates.Add.call_count == 3



import stat

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def symlink_scenario_1(state_tree):
    # Create directory structure
    dir_name = "symlink_scenario_1"
    source_dir = state_tree / dir_name
    if not source_dir.is_dir():
        source_dir.mkdir()
    source_file = source_dir / "source_file.txt"
    source_file.write_text("This is the source file...")
    symlink_file = source_dir / "symlink"
    symlink_file.symlink_to(source_file)
    yield dir_name


@pytest.fixture(scope="module")
def symlink_scenario_2(state_tree):
    # Create directory structure
    dir_name = "symlink_scenario_2"
    source_dir = state_tree / dir_name / "test"
    if not source_dir.is_dir():
        source_dir.mkdir(parents=True)
    test1 = source_dir / "test1"
    test2 = source_dir / "test2"
    test3 = source_dir / "test3"
    test_link = source_dir / "test"
    test1.touch()
    test2.touch()
    test3.touch()
    test_link.symlink_to(test3)
    yield dir_name


@pytest.fixture(scope="module")
def symlink_scenario_3(state_tree):
    # Create directory structure
    dir_name = "symlink_scenario_3"
    source_dir = state_tree / dir_name
    if not source_dir.is_dir():
        source_dir.mkdir(parents=True)
    # Create a file with the same name but is not a symlink
    source_file = source_dir / "not_a_symlink" / "symlink"
    source_file.parent.mkdir(parents=True)
    source_file.write_text("This is the source file...")
    # Create other fluff files
    just_a_file = source_dir / "just_a_file.txt"
    just_a_file.touch()
    dummy_file = source_dir / "notasymlink"
    dummy_file.touch()
    # Create symlink to source with the same name
    symlink_file = source_dir / "symlink"
    symlink_file.symlink_to(source_file)
    yield dir_name


@pytest.mark.parametrize("test", (False, True))
def test_recurse(file, tmp_path, grail, test):
    """
    file.recurse
    """
    name = tmp_path / "grail-dest-dir"
    ret = file.recurse(name=str(name), source="salt://grail", test=test)
    if test is False:
        assert ret.result is True

        scene_36_src = grail / "36" / "scene"
        scene_36_dst = name / "36" / "scene"
        assert scene_36_dst.is_file()
        assert scene_36_src.read_text() == scene_36_dst.read_text()
    else:
        assert ret.result is None

        scene_36_dst = name / "36" / "scene"
        assert scene_36_dst.is_file() is False
        assert name.exists() is False


@pytest.mark.parametrize("saltenv_param", ("__env__", "saltenv"))
def test_recurse_specific_env(file, tmp_path, holy, saltenv_param):
    """
    file.recurse passing __env__
    """
    name = tmp_path / "holy-dest-dir"
    kwargs = {saltenv_param: "prod"}
    ret = file.recurse(name=str(name), source="salt://holy", **kwargs)
    assert ret.result is True

    scene_32_src = holy / "32" / "scene"
    scene_32_dst = name / "32" / "scene"
    assert scene_32_dst.is_file()
    assert scene_32_src.read_text() == scene_32_dst.read_text()


def test_recurse_specific_env_in_url(file, tmp_path, holy):
    """
    file.recurse passing __env__
    """
    name = tmp_path / "holy-dest-dir"
    ret = file.recurse(name=str(name), source="salt://holy?saltenv=prod")
    assert ret.result is True

    scene_32_src = holy / "32" / "scene"
    scene_32_dst = name / "32" / "scene"
    assert scene_32_dst.is_file()
    assert scene_32_src.read_text() == scene_32_dst.read_text()


@pytest.mark.usefixtures("holy")
@pytest.mark.parametrize("saltenv_param", ("__env__", "saltenv"))
def test_test_recurse_specific_env(file, tmp_path, saltenv_param):
    """
    file.recurse passing __env__
    """
    name = tmp_path / "holy-dest-dir"
    kwargs = {saltenv_param: "prod"}
    ret = file.recurse(name=str(name), source="salt://holy", test=True, **kwargs)
    assert ret.result is None

    scene_32_dst = name / "32" / "scene"
    assert scene_32_dst.is_file() is False
    assert name.exists() is False


def test_recurse_template(file, tmp_path, grail):
    """
    file.recurse with jinja template enabled
    """
    name = tmp_path / "dest-dir"
    template_string = "TEMPLATE TEST STRING"
    ret = file.recurse(
        name=str(name),
        source="salt://grail",
        template="jinja",
        defaults={"spam": template_string},
    )
    assert ret.result is True

    scene_src = grail / "scene33"
    scene_dst = name / "scene33"
    assert scene_dst.is_file()
    assert scene_src.read_text() != scene_dst.read_text()
    assert template_string in scene_dst.read_text()


@pytest.mark.usefixtures("grail")
def test_recurse_clean(file, tmp_path):
    """
    file.recurse with clean=True
    """
    name = tmp_path / "dest-dir"
    name.mkdir()
    strayfile = name / "strayfile"
    strayfile.write_text("")
    scene_33_dst = name / "scene33"
    scene_36_dst = name / "36"

    # Corner cases: replacing file with a directory and vice versa
    # <name>/36 is supposed to be a directory.
    # Create a file with the same name to see if clean handles it
    scene_36_dst.write_text("")
    # <name>/scene33 is supposed to be a file.
    # Create a directory with the same name to see if clean handles it
    scene_33_dst.mkdir()
    ret = file.recurse(name=str(name), source="salt://grail", clean=True)
    assert ret.result is True
    assert strayfile.exists() is False
    assert scene_33_dst.is_dir() is False
    assert scene_33_dst.is_file()
    assert scene_36_dst.is_file() is False
    assert scene_36_dst.is_dir()
    assert scene_36_dst.joinpath("scene").is_file() is True


@pytest.mark.usefixtures("holy")
def test_recurse_clean_specific_env(file, tmp_path):
    """
    file.recurse with clean=True and saltenv=prod
    """
    name = tmp_path / "dest-dir"
    name.mkdir()
    strayfile = name / "strayfile"
    strayfile.write_text("")
    scene_34_dst = name / "scene34"
    scene_32_dst = name / "32"

    # Corner cases: replacing file with a directory and vice versa
    # <name>/32 is supposed to be a directory.
    # Create a file with the same name to see if clean handles it
    scene_32_dst.write_text("")
    # <name>/scene34 is supposed to be a file.
    # Create a directory with the same name to see if clean handles it
    scene_34_dst.mkdir()
    ret = file.recurse(name=str(name), source="salt://holy", clean=True, saltenv="prod")
    assert ret.result is True
    assert strayfile.exists() is False
    assert scene_34_dst.is_dir() is False
    assert scene_34_dst.is_file()
    assert scene_32_dst.is_file() is False
    assert scene_32_dst.is_dir()
    assert scene_32_dst.joinpath("scene").is_file() is True


@pytest.mark.skip_on_windows(reason="'dir_mode' is not supported on Windows")
def test_recurse_issue_34945(file, tmp_path, state_tree):
    """
    This tests the case where the source dir for the file.recurse state
    does not contain any files (only subdirectories), and the dir_mode is
    being managed. For a long time, this corner case resulted in the top
    level of the destination directory being created with the wrong initial
    permissions, a problem that would be corrected later on in the
    file.recurse state via running state.directory. However, the
    file.directory state only gets called when there are files to be
    managed in that directory, and when the source directory contains only
    subdirectories, the incorrectly-set initial perms would not be
    repaired.

    This was fixed in https://github.com/saltstack/salt/pull/35309

    """
    dir_mode = "2775"
    issue_dir = "issue-34945"
    src_dir = state_tree / issue_dir
    src_file = src_dir / "foo" / "bar" / "baz" / "test_file"
    src_file.parent.mkdir(mode=0o0755, parents=True)
    src_file.write_text("Hello World!\n")

    name = tmp_path / issue_dir

    ret = file.recurse(name=str(name), source=f"salt://{issue_dir}", dir_mode=dir_mode)
    assert ret.result is True
    assert name.is_dir()
    assert src_dir.stat().st_mode != name.stat().st_mode
    actual_dir_mode = oct(stat.S_IMODE(name.stat().st_mode))[-4:]
    assert actual_dir_mode == dir_mode


def test_recurse_issue_40578(file, state_tree, tmp_path):
    """
    This ensures that the state doesn't raise an exception when it
    encounters a file with a unicode filename in the process of invoking
    file.source_list.
    """
    name = tmp_path / "dst-dir"
    src_dir = state_tree / "соль"
    src_dir.mkdir()
    filenames = ("foo.txt", "спам.txt", "яйца.txt")
    for fname in filenames:
        src_dir.joinpath(fname).write_text("bar")

    ret = file.recurse(name=str(name), source="salt://соль")
    assert ret.result is True
    assert sorted(p.name for p in name.iterdir()) == sorted(filenames)


@pytest.mark.skip_on_windows(reason="Mode not available in Windows")
def test_issue_2726_mode_kwarg(modules, tmp_path, state_tree):
    dir1 = tmp_path / "dir1"
    dir2 = tmp_path / "dir2"
    with pytest.helpers.temp_file(
        "testfile", directory=state_tree, contents="test file contents"
    ), pytest.helpers.temp_file(
        "testfile",
        directory=state_tree / "testappend",
        contents="test file append contents",
    ):
        # Let's test for the wrong usage approach
        bad_mode_kwarg_testfile = dir1 / "bad_mode_kwarg" / "testfile"
        bad_template = [
            f"{bad_mode_kwarg_testfile}:",
            "  file.recurse:",
            "    - source: salt://testfile",
            "    - mode: 644",
        ]
        ret = modules.state.template_str("\n".join(bad_template))
        for state_run in ret:
            assert state_run.result is False
            assert (
                "'mode' is not allowed in 'file.recurse'. Please use 'file_mode' and 'dir_mode'."
                in state_run.comment
            )
            assert (
                "TypeError: managed() got multiple values for keyword "
                "argument 'mode'" not in state_run.comment
            )

        # Now, the correct usage approach
        good_mode_kwargs_testfile = dir2 / "good_mode_kwargs" / "testappend"
        good_template = [
            f"{good_mode_kwargs_testfile}:",
            "  file.recurse:",
            "    - source: salt://testappend",
            "    - dir_mode: 744",
            "    - file_mode: 644",
        ]
        ret = modules.state.template_str("\n".join(good_template))
        for state_run in ret:
            assert state_run.result is True


def test_issue_64630_keep_symlinks_true(file, symlink_scenario_1, tmp_path):
    """
    Make sure that symlinks are created and that there isn't an error when there
    are no conflicting target files
    """
    target_dir = tmp_path / symlink_scenario_1  # Target for the file.recurse state
    target_file = target_dir / "source_file.txt"
    target_symlink = target_dir / "symlink"

    ret = file.recurse(
        name=str(target_dir), source=f"salt://{target_dir.name}", keep_symlinks=True
    )
    assert ret.result is True

    assert target_dir.exists()
    assert target_file.is_file()
    assert target_symlink.is_symlink()


def test_issue_64630_keep_symlinks_false(file, symlink_scenario_1, tmp_path):
    """
    Make sure that symlinks are created as files and that there isn't an error
    """
    target_dir = tmp_path / symlink_scenario_1  # Target for the file.recurse state
    target_file = target_dir / "source_file.txt"
    target_symlink = target_dir / "symlink"

    ret = file.recurse(
        name=str(target_dir), source=f"salt://{target_dir.name}", keep_symlinks=False
    )
    assert ret.result is True

    assert target_dir.exists()
    assert target_file.is_file()
    assert target_symlink.is_file()
    assert target_file.read_text() == target_symlink.read_text()


def test_issue_64630_keep_symlinks_conflicting_force_symlinks_false(
    file, symlink_scenario_1, tmp_path
):
    """
    Make sure that symlinks are not created when there is a conflict. The state
    should return False
    """
    target_dir = tmp_path / symlink_scenario_1  # Target for the file.recurse state
    target_file = target_dir / "source_file.txt"
    target_symlink = target_dir / "symlink"

    # Create the conflicting file
    target_symlink.parent.mkdir(parents=True)
    target_symlink.touch()
    assert target_symlink.is_file()

    ret = file.recurse(
        name=str(target_dir),
        source=f"salt://{target_dir.name}",
        keep_symlinks=True,
        force_symlinks=False,
    )
    # We expect it to fail
    assert ret.result is False

    # And files not to be created properly
    assert target_dir.exists()
    assert target_file.is_file()
    assert target_symlink.is_file()


def test_issue_64630_keep_symlinks_conflicting_force_symlinks_true(
    file, symlink_scenario_1, tmp_path
):
    """
    Make sure that symlinks are created when there is a conflict with an
    existing file.
    """
    target_dir = tmp_path / symlink_scenario_1  # Target for the file.recurse state
    target_file = target_dir / "source_file.txt"
    target_symlink = target_dir / "symlink"

    # Create the conflicting file
    target_symlink.parent.mkdir(parents=True)
    target_symlink.touch()
    assert target_symlink.is_file()

    ret = file.recurse(
        name=str(target_dir),
        source=f"salt://{target_dir.name}",
        force_symlinks=True,
        keep_symlinks=True,
    )
    assert ret.result is True

    assert target_dir.exists()
    assert target_file.is_file()
    assert target_symlink.is_symlink()


def test_issue_64630_keep_symlinks_similar_names(file, symlink_scenario_3, tmp_path):
    """
    Make sure that symlinks are created when there is a file that shares part
    of the name of the actual symlink file. I'm not sure what I'm testing here
    as I couldn't really get this to fail either way
    """
    target_dir = tmp_path / symlink_scenario_3  # Target for the file.recurse state
    # symlink target, but has the same name as the symlink itself
    target_source = target_dir / "not_a_symlink" / "symlink"
    target_symlink = target_dir / "symlink"
    decoy_file = target_dir / "notasymlink"
    just_a_file = target_dir / "just_a_file.txt"

    ret = file.recurse(
        name=str(target_dir), source=f"salt://{target_dir.name}", keep_symlinks=True
    )
    assert ret.result is True

    assert target_dir.exists()
    assert target_source.is_file()
    assert decoy_file.is_file()
    assert just_a_file.is_file()
    assert target_symlink.is_symlink()


def test_issue_62117(file, symlink_scenario_2, tmp_path):
    target_dir = tmp_path / symlink_scenario_2 / "test"
    target_file_1 = target_dir / "test1"
    target_file_2 = target_dir / "test2"
    target_file_3 = target_dir / "test3"
    target_symlink = target_dir / "test"

    ret = file.recurse(
        name=str(target_dir),
        source=f"salt://{target_dir.parent.name}/test",
        clean=True,
        keep_symlinks=True,
    )
    assert ret.result is True

    assert target_dir.exists()
    assert target_file_1.is_file()
    assert target_file_2.is_file()
    assert target_file_3.is_file()
    assert target_symlink.is_symlink()



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.unit.modules.virtualenv_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

# Import python libraries

import sys

import salt.modules.virtualenv_mod as virtualenv_mod
from salt.exceptions import CommandExecutionError
from tests.support.helpers import ForceImportErrorOn, TstSuiteLoggingHandler
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class VirtualenvTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        base_virtualenv_mock = MagicMock()
        base_virtualenv_mock.__version__ = "1.9.1"
        patcher = patch("salt.utils.path.which", lambda exe: exe)
        patcher.start()
        self.addCleanup(patcher.stop)
        return {
            virtualenv_mod: {
                "__opts__": {"venv_bin": "virtualenv"},
                "_install_script": MagicMock(
                    return_value={
                        "retcode": 0,
                        "stdout": "Installed script!",
                        "stderr": "",
                    }
                ),
                "sys.modules": {"virtualenv": base_virtualenv_mock},
            }
        }

    def test_issue_6029_deprecated_distribute(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})

        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create(
                "/tmp/foo", system_site_packages=True, distribute=True
            )
            mock.assert_called_once_with(
                ["virtualenv", "--distribute", "--system-site-packages", "/tmp/foo"],
                runas=None,
                python_shell=False,
            )

        with TstSuiteLoggingHandler() as handler:
            # Let's fake a higher virtualenv version
            virtualenv_mock = MagicMock()
            virtualenv_mock.__version__ = "1.10rc1"
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
                with patch.dict("sys.modules", {"virtualenv": virtualenv_mock}):
                    virtualenv_mod.create(
                        "/tmp/foo", system_site_packages=True, distribute=True
                    )
                    mock.assert_called_once_with(
                        ["virtualenv", "--system-site-packages", "/tmp/foo"],
                        runas=None,
                        python_shell=False,
                    )

                # Are we logging the deprecation information?
                self.assertIn(
                    "INFO:The virtualenv '--distribute' option has been "
                    "deprecated in virtualenv(>=1.10), as such, the "
                    "'distribute' option to `virtualenv.create()` has "
                    "also been deprecated and it's not necessary anymore.",
                    handler.messages,
                )

    def test_issue_6030_deprecated_never_download(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})

        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", never_download=True)
            mock.assert_called_once_with(
                ["virtualenv", "--never-download", "/tmp/foo"],
                runas=None,
                python_shell=False,
            )

        with TstSuiteLoggingHandler() as handler:
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            # Let's fake a higher virtualenv version
            virtualenv_mock = MagicMock()
            virtualenv_mock.__version__ = "1.10rc1"
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
                with patch.dict("sys.modules", {"virtualenv": virtualenv_mock}):
                    virtualenv_mod.create("/tmp/foo", never_download=True)
                    mock.assert_called_once_with(
                        ["virtualenv", "/tmp/foo"], runas=None, python_shell=False
                    )

                # Are we logging the deprecation information?
                self.assertIn(
                    "INFO:--never-download was deprecated in 1.10.0, "
                    "but reimplemented in 14.0.0. If this feature is needed, "
                    "please install a supported virtualenv version.",
                    handler.messages,
                )

    def test_issue_6031_multiple_extra_search_dirs(self):
        extra_search_dirs = ["/tmp/bar-1", "/tmp/bar-2", "/tmp/bar-3"]

        # Passing extra_search_dirs as a list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", extra_search_dir=extra_search_dirs)
            mock.assert_called_once_with(
                [
                    "virtualenv",
                    "--extra-search-dir=/tmp/bar-1",
                    "--extra-search-dir=/tmp/bar-2",
                    "--extra-search-dir=/tmp/bar-3",
                    "/tmp/foo",
                ],
                runas=None,
                python_shell=False,
            )

        # Passing extra_search_dirs as comma separated list
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create(
                "/tmp/foo", extra_search_dir=",".join(extra_search_dirs)
            )
            mock.assert_called_once_with(
                [
                    "virtualenv",
                    "--extra-search-dir=/tmp/bar-1",
                    "--extra-search-dir=/tmp/bar-2",
                    "--extra-search-dir=/tmp/bar-3",
                    "/tmp/foo",
                ],
                runas=None,
                python_shell=False,
            )

    def test_unapplicable_options(self):
        # ----- Virtualenv using pyvenv options ----------------------------->
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                virtualenv_mod.create,
                "/tmp/foo",
                venv_bin="virtualenv",
                upgrade=True,
            )

        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                virtualenv_mod.create,
                "/tmp/foo",
                venv_bin="virtualenv",
                symlinks=True,
            )
        # <---- Virtualenv using pyvenv options ------------------------------

        # ----- pyvenv using virtualenv options ----------------------------->
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(
            virtualenv_mod.__salt__,
            {"cmd.run_all": mock, "cmd.which_bin": lambda _: "pyvenv"},
        ):
            self.assertRaises(
                CommandExecutionError,
                virtualenv_mod.create,
                "/tmp/foo",
                venv_bin="pyvenv",
                python="python2.7",
            )

        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                virtualenv_mod.create,
                "/tmp/foo",
                venv_bin="pyvenv",
                prompt="PY Prompt",
            )

        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                virtualenv_mod.create,
                "/tmp/foo",
                venv_bin="pyvenv",
                never_download=True,
            )

        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            self.assertRaises(
                CommandExecutionError,
                virtualenv_mod.create,
                "/tmp/foo",
                venv_bin="pyvenv",
                extra_search_dir="/tmp/bar",
            )
        # <---- pyvenv using virtualenv options ------------------------------

    def test_get_virtualenv_version_from_shell(self):
        with ForceImportErrorOn("virtualenv"):

            # ----- virtualenv binary not available ------------------------->
            mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(
                    CommandExecutionError,
                    virtualenv_mod.create,
                    "/tmp/foo",
                )
            # <---- virtualenv binary not available --------------------------

            # ----- virtualenv binary present but > 0 exit code ------------->
            mock = MagicMock(
                side_effect=[
                    {"retcode": 1, "stdout": "", "stderr": "This is an error"},
                    {"retcode": 0, "stdout": ""},
                ]
            )
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
                self.assertRaises(
                    CommandExecutionError,
                    virtualenv_mod.create,
                    "/tmp/foo",
                    venv_bin="virtualenv",
                )
            # <---- virtualenv binary present but > 0 exit code --------------

            # ----- virtualenv binary returns 1.9.1 as its version --------->
            mock = MagicMock(
                side_effect=[
                    {"retcode": 0, "stdout": "1.9.1"},
                    {"retcode": 0, "stdout": ""},
                ]
            )
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
                virtualenv_mod.create("/tmp/foo", never_download=True)
                mock.assert_called_with(
                    ["virtualenv", "--never-download", "/tmp/foo"],
                    runas=None,
                    python_shell=False,
                )
            # <---- virtualenv binary returns 1.9.1 as its version ----------

            # ----- virtualenv binary returns 1.10rc1 as its version ------->
            mock = MagicMock(
                side_effect=[
                    {"retcode": 0, "stdout": "1.10rc1"},
                    {"retcode": 0, "stdout": ""},
                ]
            )
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
                virtualenv_mod.create("/tmp/foo", never_download=True)
                mock.assert_called_with(
                    ["virtualenv", "/tmp/foo"], runas=None, python_shell=False
                )
            # <---- virtualenv binary returns 1.10rc1 as its version --------

    def test_python_argument(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})

        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create(
                "/tmp/foo",
                python=sys.executable,
            )
            mock.assert_called_once_with(
                ["virtualenv", f"--python={sys.executable}", "/tmp/foo"],
                runas=None,
                python_shell=False,
            )

    def test_prompt_argument(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", prompt="PY Prompt")
            mock.assert_called_once_with(
                ["virtualenv", "--prompt='PY Prompt'", "/tmp/foo"],
                runas=None,
                python_shell=False,
            )

        # Now with some quotes on the mix
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", prompt="'PY' Prompt")
            mock.assert_called_once_with(
                ["virtualenv", "--prompt=''PY' Prompt'", "/tmp/foo"],
                runas=None,
                python_shell=False,
            )

        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", prompt='"PY" Prompt')
            mock.assert_called_once_with(
                ["virtualenv", "--prompt='\"PY\" Prompt'", "/tmp/foo"],
                runas=None,
                python_shell=False,
            )

    def test_clear_argument(self):
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", clear=True)
            mock.assert_called_once_with(
                ["virtualenv", "--clear", "/tmp/foo"], runas=None, python_shell=False
            )

    def test_upgrade_argument(self):
        # We test for pyvenv only because with virtualenv this is un
        # unsupported option.
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", venv_bin="pyvenv", upgrade=True)
            mock.assert_called_once_with(
                ["pyvenv", "--upgrade", "/tmp/foo"], runas=None, python_shell=False
            )

    def test_symlinks_argument(self):
        # We test for pyvenv only because with virtualenv this is un
        # unsupported option.
        mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock}):
            virtualenv_mod.create("/tmp/foo", venv_bin="pyvenv", symlinks=True)
            mock.assert_called_once_with(
                ["pyvenv", "--symlinks", "/tmp/foo"], runas=None, python_shell=False
            )

    def test_virtualenv_ver(self):
        """
        test virtualenv_ver when there is no ImportError
        """
        ret = virtualenv_mod.virtualenv_ver(venv_bin="pyvenv")
        assert ret == (1, 9, 1)

    def test_virtualenv_ver_importerror(self):
        """
        test virtualenv_ver when there is an ImportError
        """
        with ForceImportErrorOn("virtualenv"):
            mock_ver = MagicMock(return_value={"retcode": 0, "stdout": "1.9.1"})
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock_ver}):
                ret = virtualenv_mod.virtualenv_ver(venv_bin="pyenv")
        assert ret == (1, 9, 1)

    def test_virtualenv_ver_importerror_cmd_error(self):
        """
        test virtualenv_ver when there is an ImportError
        and virtualenv --version does not return anything
        """
        with ForceImportErrorOn("virtualenv"):
            mock_ver = MagicMock(return_value={"retcode": 0, "stdout": ""})
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock_ver}):
                with self.assertRaises(CommandExecutionError):
                    virtualenv_mod.virtualenv_ver(venv_bin="pyenv")

    def test_virtualenv_importerror_ver_output(self):
        """
        test virtualenv_ver when there is an ImportError
        and virtualenv --version returns the various
        --versions outputs
        """
        stdout = (
            ("1.9.2", (1, 9, 2)),
            ("1.9rc2", (1, 9)),
            (
                "virtualenv 20.0.0 from"
                " /home/ch3ll/.pyenv/versions/3.6.4/envs/virtualenv/lib/python3.6/site-packages/virtualenv/__init__.py",
                (20, 0, 0),
            ),
            ("16.7.10", (16, 7, 10)),
        )
        for stdout, expt in stdout:
            with ForceImportErrorOn("virtualenv"):
                mock_ver = MagicMock(return_value={"retcode": 0, "stdout": stdout})
                with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock_ver}):
                    ret = virtualenv_mod.virtualenv_ver(venv_bin="pyenv")
                    assert ret == expt

    def test_issue_57734_debian_package(self):
        virtualenv_mock = MagicMock()
        virtualenv_mock.__version__ = "20.0.23+ds"
        with patch.dict("sys.modules", {"virtualenv": virtualenv_mock}):
            ret = virtualenv_mod.virtualenv_ver(venv_bin="pyenv")
        self.assertEqual(ret, (20, 0, 23))

    def test_issue_57734_debian_package_importerror(self):
        with ForceImportErrorOn("virtualenv"):
            mock_ver = MagicMock(
                return_value={
                    "retcode": 0,
                    "stdout": (
                        "virtualenv 20.0.23+ds from "
                        "/usr/lib/python3/dist-packages/virtualenv/__init__.py"
                    ),
                }
            )
            with patch.dict(virtualenv_mod.__salt__, {"cmd.run_all": mock_ver}):
                ret = virtualenv_mod.virtualenv_ver(venv_bin="pyenv")
        self.assertEqual(ret, (20, 0, 23))



"""
    Tests for salt.modules.boto3_route53
"""

import random
import string

import pytest

import salt.loader
import salt.modules.boto3_route53 as boto3_route53
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

try:
    import boto3

    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False

# the boto3_route53 module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
REQUIRED_BOTO3_VERSION = "1.2.1"

pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
]


def __virtual__():
    """
    Returns True/False boolean depending on if Boto3 is installed and correct
    version.
    """
    if not HAS_BOTO3:
        return False
    if Version(boto3.__version__) < Version(REQUIRED_BOTO3_VERSION):
        return (
            False,
            "The boto3 module must be greater or equal to version {}".format(
                REQUIRED_BOTO3_VERSION
            ),
        )
    return True


REGION = "us-east-1"
ACCESS_KEY = "GKTADJGHEIQSXMKKRBJ08H"
SECRET_KEY = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
CONN_PARAMETERS = {
    "region": REGION,
    "key": ACCESS_KEY,
    "keyid": SECRET_KEY,
    "profile": {},
}

LIST_RESOURCE_RECORD_SETS_RETURN = {
    "IsTruncated": True,
    "MaxItems": "100",
    "NextRecordName": "blog3.saltstack.furniture.",
    "NextRecordType": "CNAME",
    "ResourceRecordSets": [
        {
            "Name": "blog.saltstack.furniture.",
            "ResourceRecords": [{"Value": "127.0.0.1"}],
            "TTL": 60,
            "Type": "A",
        },
        {
            "Name": "blog2.saltstack.furniture.",
            "ResourceRecords": [{"Value": "127.0.0.1"}],
            "TTL": 60,
            "Type": "A",
        },
    ],
}


@pytest.mark.skipif(HAS_BOTO3 is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    Version(boto3.__version__) < Version(REQUIRED_BOTO3_VERSION),
    reason="The boto3 module must be greater or equal to version {}".format(
        REQUIRED_BOTO3_VERSION
    ),
)
class Boto3Route53TestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.boto3_route53 moodule
    """

    conn = None

    def setup_loader_modules(self):
        self.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        utils = salt.loader.utils(
            self.opts,
            whitelist=["boto3", "args", "systemd", "path", "platform"],
            context={},
        )
        return {boto3_route53: {"__utils__": utils}}

    def setUp(self):
        super().setUp()
        boto3_route53.__init__(self.opts)
        del self.opts

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        CONN_PARAMETERS["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()
        session_instance = mock_session.return_value
        session_instance.configure_mock(client=MagicMock(return_value=self.conn))
        self.paginator = MagicMock()
        self.addCleanup(delattr, self, "paginator")
        self.conn.configure_mock(get_paginator=MagicMock(return_value=self.paginator))

    def test_get_resource_records(self):
        """
        Test get_resource_records behaviour.
        """

        # The patch below is not neccesary per se,
        # as .exists returns positive as long as no exception is raised.
        with patch.object(
            self.conn,
            "list_resource_record_sets",
            return_value=LIST_RESOURCE_RECORD_SETS_RETURN,
        ):
            self.assertEqual(
                boto3_route53.get_resource_records(
                    HostedZoneId="Z2P70J7EXAMPLE",
                    StartRecordName="blog.saltstack.furniture.",
                    StartRecordType="A",
                    **CONN_PARAMETERS
                ),
                [
                    {
                        "Name": "blog.saltstack.furniture.",
                        "ResourceRecords": [{"Value": "127.0.0.1"}],
                        "TTL": 60,
                        "Type": "A",
                    }
                ],
            )



import sys

import salt.payload


def _trim_dict_in_dict(data, max_val_size, replace_with):
    """
    Takes a dictionary, max_val_size and replace_with
    and recursively loops through and replaces any values
    that are greater than max_val_size.
    """
    for key in data:
        if isinstance(data[key], dict):
            _trim_dict_in_dict(data[key], max_val_size, replace_with)
        else:
            if sys.getsizeof(data[key]) > max_val_size:
                data[key] = replace_with


def trim_dict(
    data,
    max_dict_bytes,
    percent=50.0,
    stepper_size=10,
    replace_with="VALUE_TRIMMED",
    is_msgpacked=False,
    use_bin_type=False,
):
    """
    Takes a dictionary and iterates over its keys, looking for
    large values and replacing them with a trimmed string.

    If after the first pass over dictionary keys, the dictionary
    is not sufficiently small, the stepper_size will be increased
    and the dictionary will be rescanned. This allows for progressive
    scanning, removing large items first and only making additional
    passes for smaller items if necessary.

    This function uses msgpack to calculate the size of the dictionary
    in question. While this might seem like unnecessary overhead, a
    data structure in python must be serialized in order for sys.getsizeof()
    to accurately return the items referenced in the structure.

    Ex:
    >>> salt.utils.dicttrim.trim_dict({'a': 'b', 'c': 'x' * 10000}, 100)
    {'a': 'b', 'c': 'VALUE_TRIMMED'}

    To improve performance, it is adviseable to pass in msgpacked
    data structures instead of raw dictionaries. If a msgpack
    structure is passed in, it will not be unserialized unless
    necessary.

    If a msgpack is passed in, it will be repacked if necessary
    before being returned.

    :param use_bin_type: Set this to true if "is_msgpacked=True"
                         and the msgpack data has been encoded
                         with "use_bin_type=True". This also means
                         that the msgpack data should be decoded with
                         "encoding='utf-8'".
    """
    if is_msgpacked:
        dict_size = sys.getsizeof(data)
    else:
        dict_size = sys.getsizeof(salt.payload.dumps(data))
    if dict_size > max_dict_bytes:
        if is_msgpacked:
            if use_bin_type:
                data = salt.payload.loads(data, encoding="utf-8")
            else:
                data = salt.payload.loads(data)
        while True:
            percent = float(percent)
            max_val_size = float(max_dict_bytes * (percent / 100))
            try:
                for key in data:
                    if isinstance(data[key], dict):
                        _trim_dict_in_dict(data[key], max_val_size, replace_with)
                    else:
                        if sys.getsizeof(data[key]) > max_val_size:
                            data[key] = replace_with
                percent = percent - stepper_size
                max_val_size = float(max_dict_bytes * (percent / 100))
                if use_bin_type:
                    dump_data = salt.payload.dumps(data, use_bin_type=True)
                else:
                    dump_data = salt.payload.dumps(data)
                cur_dict_size = sys.getsizeof(dump_data)
                if cur_dict_size < max_dict_bytes:
                    if is_msgpacked:  # Repack it
                        return dump_data
                    else:
                        return data
                elif max_val_size == 0:
                    if is_msgpacked:
                        return dump_data
                    else:
                        return data
            except ValueError:
                pass
        if is_msgpacked:
            if use_bin_type:
                return salt.payload.dumps(data, use_bin_type=True)
            else:
                return salt.payload.dumps(data)
        else:
            return data
    else:
        return data



"""
tests.pytests.functional.utils.test_process
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Test salt's process utility module
"""

import os
import pathlib
import time

import pytest

import salt.utils.process


class Process(salt.utils.process.SignalHandlingProcess):
    def run(self):
        pass


@pytest.fixture
def process_manager():
    _process_manager = salt.utils.process.ProcessManager(wait_for_kill=5)
    try:
        yield _process_manager
    finally:
        _process_manager.terminate()


@pytest.mark.skipif(
    "grains['osfinger'] == 'Rocky Linux-8' and grains['osarch'] == 'aarch64'",
    reason="Temporarily skip on Rocky Linux 8 Arm64",
)
def test_process_manager_60749(process_manager):
    """
    Regression test for issue #60749
    """

    process_manager.add_process(Process)
    process_manager.check_children()


def _get_num_fds(pid):
    "Determine the number of open fds for a process, linux only."
    return len(list(pathlib.Path(f"/proc/{pid}/fd").iterdir()))


@pytest.mark.skip_unless_on_linux
def test_subprocess_list_fds():
    pid = os.getpid()
    process_list = salt.utils.process.SubprocessList()

    before_num = _get_num_fds(pid)

    def target():
        pass

    process = salt.utils.process.SignalHandlingProcess(target=target)
    process.start()

    process_list.add(process)
    time.sleep(0.3)

    num = _get_num_fds(pid)
    assert num == before_num + 2
    start = time.time()
    while time.time() - start < 1:
        process_list.cleanup()
        if not process_list.processes:
            break
    assert len(process_list.processes) == 0
    assert _get_num_fds(pid) == num - 2



"""
    :codeauthor: Lukas Raska <lukas@raska.me>
"""

import pytest

import salt.utils.dictdiffer as dictdiffer
from salt.exceptions import CommandExecutionError
from salt.states import elasticsearch
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        elasticsearch: {
            "__opts__": {"test": False},
            "__utils__": {"dictdiffer.deep_diff": dictdiffer.deep_diff},
        }
    }


def test_index_absent():
    """
    Test to manage a elasticsearch index.
    """
    name = "foo"

    ret = {
        "name": name,
        "result": True,
        "comment": "Index foo is already absent",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            None,
            {name: {"test": "key"}},
            {name: {}},
            {name: {"test": "key"}},
            CommandExecutionError,
            {name: {"test": "key"}},
        ]
    )
    mock_delete = MagicMock(side_effect=[True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.index_get": mock_get,
            "elasticsearch.index_delete": mock_delete,
        },
    ):
        assert elasticsearch.index_absent(name) == ret

        ret.update(
            {
                "comment": "Successfully removed index foo",
                "changes": {"old": {"test": "key"}},
            }
        )
        assert elasticsearch.index_absent(name) == ret

        ret.update(
            {
                "comment": "Failed to remove index foo for unknown reasons",
                "result": False,
                "changes": {},
            }
        )
        assert elasticsearch.index_absent(name) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Index foo will be removed",
                    "result": None,
                    "changes": {"old": {"test": "key"}},
                }
            )
            assert elasticsearch.index_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_absent(name) == ret


def test_index_present():
    """
    Test to manage a elasticsearch index.
    """
    name = "foo"

    ret = {
        "name": name,
        "result": True,
        "comment": "Index foo is already present",
        "changes": {},
    }

    mock_exists = MagicMock(
        side_effect=[True, False, False, False, CommandExecutionError, False, False]
    )
    mock_get = MagicMock(side_effect=[{name: {"test": "key"}}, CommandExecutionError])
    mock_create = MagicMock(side_effect=[True, False, CommandExecutionError, True])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.index_get": mock_get,
            "elasticsearch.index_exists": mock_exists,
            "elasticsearch.index_create": mock_create,
        },
    ):
        assert elasticsearch.index_present(name) == ret

        ret.update(
            {
                "comment": "Successfully created index foo",
                "changes": {"new": {"test": "key"}},
            }
        )
        assert elasticsearch.index_present(name) == ret

        ret.update(
            {
                "comment": "Cannot create index foo, False",
                "result": False,
                "changes": {},
            }
        )
        assert elasticsearch.index_present(name) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Index foo does not exist and will be created",
                    "result": None,
                    "changes": {"new": {"test2": "key"}},
                }
            )
            assert elasticsearch.index_present(name, {"test2": "key"}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_absent(name) == ret


def test_alias_absent():
    """
    Test to manage a elasticsearch alias.
    """
    name = "foo"
    index = "bar"

    alias = {index: {"aliases": {name: {"test": "key"}}}}

    ret = {
        "name": name,
        "result": True,
        "comment": "Alias foo for index bar is already absent",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            None,
            {"foo2": {}},
            alias,
            alias,
            alias,
            CommandExecutionError,
            alias,
        ]
    )
    mock_delete = MagicMock(side_effect=[True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.alias_get": mock_get,
            "elasticsearch.alias_delete": mock_delete,
        },
    ):
        assert elasticsearch.alias_absent(name, index) == ret
        assert elasticsearch.alias_absent(name, index) == ret

        ret.update(
            {
                "comment": "Successfully removed alias foo for index bar",
                "changes": {"old": {"test": "key"}},
            }
        )
        assert elasticsearch.alias_absent(name, index) == ret

        ret.update(
            {
                "comment": (
                    "Failed to remove alias foo for index bar for unknown reasons"
                ),
                "result": False,
                "changes": {},
            }
        )
        assert elasticsearch.alias_absent(name, index) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Alias foo for index bar will be removed",
                    "result": None,
                    "changes": {"old": {"test": "key"}},
                }
            )
            assert elasticsearch.alias_absent(name, index) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.alias_absent(name, index) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.alias_absent(name, index) == ret


def test_alias_present():
    """
    Test to manage a elasticsearch alias.
    """
    name = "foo"
    index = "bar"

    alias = {index: {"aliases": {name: {"test": "key"}}}}

    ret = {
        "name": name,
        "result": True,
        "comment": "Alias foo for index bar is already present",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            alias,
            alias,
            None,
            None,
            None,
            alias,
            CommandExecutionError,
            None,
        ]
    )
    mock_create = MagicMock(side_effect=[True, True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.alias_get": mock_get,
            "elasticsearch.alias_create": mock_create,
        },
    ):
        assert elasticsearch.alias_present(name, index, {"test": "key"}) == ret

        ret.update(
            {
                "comment": "Successfully replaced alias foo for index bar",
                "changes": {"old": {"test": "key"}, "new": {"test2": "key"}},
            }
        )
        assert elasticsearch.alias_present(name, index, {"test2": "key"}) == ret

        ret.update(
            {
                "comment": "Successfully created alias foo for index bar",
                "changes": {"new": {"test2": "key"}},
            }
        )
        assert elasticsearch.alias_present(name, index, {"test2": "key"}) == ret

        ret.update(
            {
                "comment": "Cannot create alias foo for index bar, False",
                "result": False,
            }
        )
        assert elasticsearch.alias_present(name, index, {"test2": "key"}) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": (
                        "Alias foo for index bar does not exist and will be created"
                    ),
                    "result": None,
                    "changes": {"new": {"test2": "key"}},
                }
            )
            assert elasticsearch.alias_present(name, index, {"test2": "key"}) == ret

            ret.update(
                {
                    "comment": (
                        "Alias foo for index bar exists with wrong configuration and"
                        " will be overridden"
                    ),
                    "result": None,
                    "changes": {"old": {"test": "key"}, "new": {"test2": "key"}},
                }
            )
            assert elasticsearch.alias_present(name, index, {"test2": "key"}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.alias_present(name, index) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.alias_present(name, index) == ret


def test_index_template_absent():
    """
    Test to manage a elasticsearch index template.
    """
    name = "foo"

    index_template = {name: {"test": "key"}}

    ret = {
        "name": name,
        "result": True,
        "comment": "Index template foo is already absent",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            None,
            {"bar": {}},
            index_template,
            index_template,
            index_template,
            CommandExecutionError,
            index_template,
        ]
    )
    mock_delete = MagicMock(side_effect=[True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.index_template_get": mock_get,
            "elasticsearch.index_template_delete": mock_delete,
        },
    ):
        assert elasticsearch.index_template_absent(name) == ret
        assert elasticsearch.index_template_absent(name) == ret

        ret.update(
            {
                "comment": "Successfully removed index template foo",
                "changes": {"old": {"test": "key"}},
            }
        )
        assert elasticsearch.index_template_absent(name) == ret

        ret.update(
            {
                "comment": "Failed to remove index template foo for unknown reasons",
                "result": False,
                "changes": {},
            }
        )
        assert elasticsearch.index_template_absent(name) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Index template foo will be removed",
                    "result": None,
                    "changes": {"old": {"test": "key"}},
                }
            )
            assert elasticsearch.index_template_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_template_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_template_absent(name) == ret


def test_index_template_present():
    """
    Test to manage a elasticsearch index template.
    """
    name = "foo"

    index_template = {name: {"test": "key"}}

    ret = {
        "name": name,
        "result": True,
        "comment": "Index template foo is already present",
        "changes": {},
    }

    mock_exists = MagicMock(
        side_effect=[True, False, False, False, CommandExecutionError, False, False]
    )
    mock_create = MagicMock(side_effect=[True, False, CommandExecutionError, True])
    mock_get = MagicMock(side_effect=[index_template, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.index_template_get": mock_get,
            "elasticsearch.index_template_create": mock_create,
            "elasticsearch.index_template_exists": mock_exists,
        },
    ):
        assert elasticsearch.index_template_present(name, {"test2": "key"}) == ret

        ret.update(
            {
                "comment": "Successfully created index template foo",
                "changes": {"new": {"test": "key"}},
            }
        )
        assert elasticsearch.index_template_present(name, {"test2": "key"}) == ret

        ret.update(
            {
                "comment": "Cannot create index template foo, False",
                "result": False,
                "changes": {},
            }
        )
        assert elasticsearch.index_template_present(name, {"test2": "key"}) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Index template foo does not exist and will be created",
                    "result": None,
                    "changes": {"new": {"test2": "key"}},
                }
            )
            assert elasticsearch.index_template_present(name, {"test2": "key"}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_template_present(name, {}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_template_present(name, {}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.index_template_present(name, {}) == ret


def test_index_template_present_check_definition():
    """
    Test to manage a elasticsearch index template.
    with check_definition set
    """
    name = "foo"

    index_template = {
        name: {"test2": "key", "aliases": {}, "mappings": {}, "settings": {}}
    }

    expected = {
        "name": name,
        "result": True,
        "comment": "Index template foo is already present and up to date",
        "changes": {},
    }

    mock_exists = MagicMock(side_effect=[True])
    mock_create = MagicMock(side_effect=[True])
    mock_get = MagicMock(side_effect=[index_template])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.index_template_get": mock_get,
            "elasticsearch.index_template_create": mock_create,
            "elasticsearch.index_template_exists": mock_exists,
        },
    ):

        ret = elasticsearch.index_template_present(
            name, {"test2": "key", "aliases": {}}, check_definition=True
        )
        assert expected == ret


def test_index_template_present_check_definition_alias_not_empty():
    """
    Test to manage a elasticsearch index template.
    with check_definition set and alias is not empty
    """
    name = "foo"

    index_template = {
        name: {"test2": "key", "aliases": {}, "mappings": {}, "settings": {}}
    }

    expected = {
        "name": name,
        "result": True,
        "comment": "Successfully updated index template foo",
        "changes": {"new": {"aliases": {"alias1": {}}}, "old": {"aliases": {}}},
    }

    mock_exists = MagicMock(side_effect=[True])
    mock_create = MagicMock(side_effect=[True])
    mock_get = MagicMock(side_effect=[index_template])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.index_template_get": mock_get,
            "elasticsearch.index_template_create": mock_create,
            "elasticsearch.index_template_exists": mock_exists,
        },
    ):

        ret = elasticsearch.index_template_present(
            name, {"test2": "key", "aliases": {"alias1": {}}}, check_definition=True
        )
        assert expected == ret


def test_pipeline_absent():
    """
    Test to manage a elasticsearch pipeline.
    """
    name = "foo"

    pipeline = {name: {"test": "key"}}

    ret = {
        "name": name,
        "result": True,
        "comment": "Pipeline foo is already absent",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            None,
            {"foo2": {}},
            pipeline,
            pipeline,
            pipeline,
            CommandExecutionError,
            pipeline,
        ]
    )
    mock_delete = MagicMock(side_effect=[True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.pipeline_get": mock_get,
            "elasticsearch.pipeline_delete": mock_delete,
        },
    ):
        assert elasticsearch.pipeline_absent(name) == ret
        assert elasticsearch.pipeline_absent(name) == ret

        ret.update(
            {
                "comment": "Successfully removed pipeline foo",
                "changes": {"old": {"test": "key"}},
            }
        )
        assert elasticsearch.pipeline_absent(name) == ret

        ret.update(
            {
                "comment": "Failed to remove pipeline foo for unknown reasons",
                "result": False,
                "changes": {},
            }
        )
        assert elasticsearch.pipeline_absent(name) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Pipeline foo will be removed",
                    "result": None,
                    "changes": {"old": {"test": "key"}},
                }
            )
            assert elasticsearch.pipeline_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.pipeline_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.pipeline_absent(name) == ret


def test_pipeline_present():
    """
    Test to manage a elasticsearch pipeline.
    """
    name = "foo"

    pipeline = {name: {"test": "key"}}

    ret = {
        "name": name,
        "result": True,
        "comment": "Pipeline foo is already present",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            pipeline,
            pipeline,
            None,
            None,
            None,
            pipeline,
            CommandExecutionError,
            None,
        ]
    )
    mock_create = MagicMock(side_effect=[True, True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.pipeline_get": mock_get,
            "elasticsearch.pipeline_create": mock_create,
        },
    ):
        assert elasticsearch.pipeline_present(name, {"test": "key"}) == ret

        ret.update(
            {
                "comment": "Successfully replaced pipeline foo",
                "changes": {"old": {"test": "key"}, "new": {"test2": "key"}},
            }
        )
        assert elasticsearch.pipeline_present(name, {"test2": "key"}) == ret

        ret.update(
            {
                "comment": "Successfully created pipeline foo",
                "changes": {"new": {"test2": "key"}},
            }
        )
        assert elasticsearch.pipeline_present(name, {"test2": "key"}) == ret

        ret.update({"comment": "Cannot create pipeline foo, False", "result": False})
        assert elasticsearch.pipeline_present(name, {"test2": "key"}) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Pipeline foo does not exist and will be created",
                    "result": None,
                    "changes": {"new": {"test2": "key"}},
                }
            )
            assert elasticsearch.pipeline_present(name, {"test2": "key"}) == ret

            ret.update(
                {
                    "comment": (
                        "Pipeline foo exists with wrong configuration and will be"
                        " overridden"
                    ),
                    "result": None,
                    "changes": {"old": {"test": "key"}, "new": {"test2": "key"}},
                }
            )
            assert elasticsearch.pipeline_present(name, {"test2": "key"}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.pipeline_present(name, {}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.pipeline_present(name, {}) == ret


def test_search_template_absent():
    """
    Test to manage a elasticsearch search template.
    """
    name = "foo"

    template = {"template": '{"test": "key"}'}

    ret = {
        "name": name,
        "result": True,
        "comment": "Search template foo is already absent",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            None,
            template,
            template,
            template,
            CommandExecutionError,
            template,
        ]
    )
    mock_delete = MagicMock(side_effect=[True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.search_template_get": mock_get,
            "elasticsearch.search_template_delete": mock_delete,
        },
    ):
        assert elasticsearch.search_template_absent(name) == ret

        ret.update(
            {
                "comment": "Successfully removed search template foo",
                "changes": {"old": {"test": "key"}},
            }
        )
        assert elasticsearch.search_template_absent(name) == ret

        ret.update(
            {
                "comment": "Failed to remove search template foo for unknown reasons",
                "result": False,
                "changes": {},
            }
        )
        assert elasticsearch.search_template_absent(name) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Search template foo will be removed",
                    "result": None,
                    "changes": {"old": {"test": "key"}},
                }
            )
            assert elasticsearch.search_template_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.search_template_absent(name) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.search_template_absent(name) == ret


def test_search_template_present():
    """
    Test to manage a elasticsearch search template.
    """
    name = "foo"

    template = {"template": '{"test": "key"}'}

    ret = {
        "name": name,
        "result": True,
        "comment": "Search template foo is already present",
        "changes": {},
    }

    mock_get = MagicMock(
        side_effect=[
            template,
            template,
            None,
            None,
            None,
            template,
            CommandExecutionError,
            None,
        ]
    )
    mock_create = MagicMock(side_effect=[True, True, False, CommandExecutionError])

    with patch.dict(
        elasticsearch.__salt__,
        {
            "elasticsearch.search_template_get": mock_get,
            "elasticsearch.search_template_create": mock_create,
        },
    ):
        assert elasticsearch.search_template_present(name, {"test": "key"}) == ret

        ret.update(
            {
                "comment": "Successfully replaced search template foo",
                "changes": {"old": {"test": "key"}, "new": {"test2": "key"}},
            }
        )
        assert elasticsearch.search_template_present(name, {"test2": "key"}) == ret

        ret.update(
            {
                "comment": "Successfully created search template foo",
                "changes": {"new": {"test2": "key"}},
            }
        )
        assert elasticsearch.search_template_present(name, {"test2": "key"}) == ret

        ret.update(
            {"comment": "Cannot create search template foo, False", "result": False}
        )
        assert elasticsearch.search_template_present(name, {"test2": "key"}) == ret

        with patch.dict(elasticsearch.__opts__, {"test": True}):
            ret.update(
                {
                    "comment": "Search template foo does not exist and will be created",
                    "result": None,
                    "changes": {"new": {"test2": "key"}},
                }
            )
            assert elasticsearch.search_template_present(name, {"test2": "key"}) == ret

            ret.update(
                {
                    "comment": (
                        "Search template foo exists with wrong configuration and will"
                        " be overridden"
                    ),
                    "result": None,
                    "changes": {"old": {"test": "key"}, "new": {"test2": "key"}},
                }
            )
            assert elasticsearch.search_template_present(name, {"test2": "key"}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.search_template_present(name, {}) == ret

        ret.update({"comment": "", "result": False, "changes": {}})
        assert elasticsearch.search_template_present(name, {}) == ret



from collections import OrderedDict

import pytest
from jinja2.exceptions import TemplateNotFound

import salt.utils.jinja


def test_utils_jinja_cache_removed_file_from_root(temp_salt_minion, tmp_path):
    """
    this tests for a condition where an included jinja template
    is removed from the salt filesystem, but is still loaded from
    the cache.
    """
    opts = temp_salt_minion.config.copy()
    file_root = tmp_path / "root"
    file_root.mkdir(parents=True, exist_ok=True)
    cache_root = tmp_path / "cache"
    cache_root.mkdir(parents=True, exist_ok=True)
    filename = "jinja_cache"
    sls_file = file_root / f"{filename}.sls"
    jinja_file = file_root / f"{filename}.jinja"
    sls_file.write_text("{% include '" + filename + ".jinja' %}")
    jinja_file.write_text("{% set this = 'that' %}")

    # Stop using OrderedDict once we drop Py3.5 support
    opts["file_roots"] = OrderedDict()
    opts["file_roots"]["base"] = [str(file_root)]
    opts["cachedir"] = str(cache_root)
    opts["master_type"] = "disable"
    opts["file_client"] = "local"

    loader = salt.utils.jinja.SaltCacheLoader(
        opts,
        "base",
    )
    # sls file is in the root
    loader.get_source(None, sls_file.name)
    # jinja file is in the root
    loader.get_source(None, jinja_file.name)
    # both files cached
    assert len(loader.cached) == 2

    # remove the jinja file and reset loader
    jinja_file.unlink()
    loader = salt.utils.jinja.SaltCacheLoader(
        opts,
        "base",
    )
    # sls file is still in the root
    loader.get_source(None, sls_file.name)
    # jinja file is gone from the root
    with pytest.raises(TemplateNotFound):
        loader.get_source(None, jinja_file.name)
    # only one was cached this run
    assert len(loader.cached) == 1
    # the cached jinja file is still present, but not used
    assert (cache_root / "files" / "base" / jinja_file.name).exists()



"""
    :synopsis: Unit Tests for 'module.yumkernelpkg'
    :platform: Linux
    :maturity: develop
    .. versionadded:: 2018.3.0
"""

# pylint: disable=invalid-name,no-member

import pytest

from salt.exceptions import CommandExecutionError
from tests.support.kernelpkg import KernelPkgTestCase
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

try:
    import salt.modules.kernelpkg_linux_yum as kernelpkg
    import salt.modules.yumpkg as pkg

    HAS_MODULES = True
except ImportError:
    HAS_MODULES = False


@pytest.mark.skipif(not HAS_MODULES, reason="Salt modules could not be loaded")
class YumKernelPkgTestCase(KernelPkgTestCase, TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.kernelpkg_linux_yum
    """

    _kernelpkg = kernelpkg
    KERNEL_LIST = ["3.10.0-327.el7", "3.11.0-327.el7", "4.9.1-100.el7"]
    LATEST = KERNEL_LIST[-1]
    OS_ARCH = "x86_64"
    OS_NAME = "RedHat"
    OS_MAJORRELEASE = "7"

    def setup_loader_modules(self):
        return {
            kernelpkg: {
                "__grains__": {
                    "os": self.OS_NAME,
                    "osmajorrelease": self.OS_MAJORRELEASE,
                    "kernelrelease": f"{self.KERNEL_LIST[0]}.{self.OS_ARCH}",
                },
                "__salt__": {
                    "pkg.normalize_name": pkg.normalize_name,
                    "pkg.upgrade": MagicMock(return_value={}),
                    "pkg.list_pkgs": MagicMock(return_value={}),
                    "pkg.version": MagicMock(return_value=self.KERNEL_LIST),
                    "system.reboot": MagicMock(return_value=None),
                    "config.get": MagicMock(return_value=True),
                },
            },
            pkg: {"__grains__": {"osarch": self.OS_ARCH}},
        }

    def test_list_installed(self):
        """
        Test - Return the latest installed kernel version
        """
        mock = MagicMock(return_value=self.KERNEL_LIST)
        with patch.dict(self._kernelpkg.__salt__, {"pkg.version": mock}):
            self.assertListEqual(self._kernelpkg.list_installed(), self.KERNEL_LIST)

    def test_list_installed_none(self):
        """
        Test - Return the latest installed kernel version
        """
        mock = MagicMock(return_value=None)
        with patch.dict(self._kernelpkg.__salt__, {"pkg.version": mock}):
            self.assertListEqual(self._kernelpkg.list_installed(), [])

    def test_remove_success(self):
        """
        Test - remove kernel package
        """
        mock = MagicMock(return_value={"retcode": 0, "stderr": []})
        with patch.dict(self._kernelpkg.__salt__, {"cmd.run_all": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]
            ):
                with patch.object(
                    self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
                ):
                    result = self._kernelpkg.remove(release=self.KERNEL_LIST[0])
                    self.assertIn("removed", result)
                    target = "{}-{}".format(
                        self._kernelpkg._package_name(), self.KERNEL_LIST[0]
                    )  # pylint: disable=protected-access
                    self.assertListEqual(result["removed"], [target])

    def test_remove_error(self):
        """
        Test - remove kernel package
        """
        mock = MagicMock(return_value={"retcode": -1, "stderr": []})
        with patch.dict(self._kernelpkg.__salt__, {"cmd.run_all": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]
            ):
                with patch.object(
                    self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
                ):
                    self.assertRaises(
                        CommandExecutionError,
                        self._kernelpkg.remove,
                        release=self.KERNEL_LIST[0],
                    )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.serverdensity_device as serverdensity_device
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {serverdensity_device: {}}


def test_monitored():
    """
    Test to device is monitored with Server Density.
    """
    name = "my_special_server"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock_dict = MagicMock(return_value={"id": name})
    mock_t = MagicMock(side_effect=[True, {"agentKey": True}, [{"agentKey": True}]])
    mock_sd = MagicMock(side_effect=[["sd-agent"], [], True])
    with patch.multiple(
        serverdensity_device,
        __salt__={
            "status.all_status": mock_dict,
            "grains.items": mock_dict,
            "serverdensity_device.ls": mock_t,
            "pkg.list_pkgs": mock_sd,
            "serverdensity_device.install_agent": mock_sd,
        },
        __opts__={"test": False},
    ):
        comt = (
            "Such server name already exists in this"
            " Server Density account. And sd-agent is installed"
        )
        ret.update({"comment": comt})
        assert serverdensity_device.monitored(name) == ret

        comt = "Successfully installed agent and created device in Server Density db."
        ret.update(
            {
                "comment": comt,
                "changes": {
                    "created_device": {"agentKey": True},
                    "installed_agent": True,
                },
            }
        )
        assert serverdensity_device.monitored(name) == ret



"""
    :codeauthor: Gareth J. Greenaway <ggreenaway@vmware.com>
"""

import os
import textwrap

import pytest

import salt.modules.pkg_resource as pkg_resource
import salt.modules.zypperpkg as zypper
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        zypper: {
            "rpm": None,
            "_systemd_scope": MagicMock(return_value=False),
            "osrelease_info": [15, 3],
            "__salt__": {"pkg_resource.parse_targets": pkg_resource.parse_targets},
        },
        pkg_resource: {"__grains__": {"os": "SUSE"}},
    }


def test_list_pkgs_no_context():
    """
    Test packages listing.

    :return:
    """

    def _add_data(data, key, value):
        data.setdefault(key, []).append(value)

    rpm_out = [
        "protobuf-java_|-(none)_|-2.6.1_|-3.1.develHead_|-noarch_|-(none)_|-1499257756",
        "yast2-ftp-server_|-(none)_|-3.1.8_|-8.1_|-x86_64_|-(none)_|-1499257798",
        "jose4j_|-(none)_|-0.4.4_|-2.1.develHead_|-noarch_|-(none)_|-1499257756",
        "apache-commons-cli_|-(none)_|-1.2_|-1.233_|-noarch_|-(none)_|-1498636510",
        "jakarta-commons-discovery_|-(none)_|-0.4_|-129.686_|-noarch_|-(none)_|-1498636511",
        "susemanager-build-keys-web_|-(none)_|-12.0_|-5.1.develHead_|-noarch_|-(none)_|-1498636510",
        "gpg-pubkey_|-(none)_|-39db7c82_|-5847eb1f_|-(none)_|-(none)_|-1519203802",
        "gpg-pubkey_|-(none)_|-8a7c64f9_|-5aaa93ca_|-(none)_|-(none)_|-1529925595",
        "kernel-default_|-(none)_|-4.4.138_|-94.39.1_|-x86_64_|-(none)_|-1529936067",
        "kernel-default_|-(none)_|-4.4.73_|-5.1_|-x86_64_|-(none)_|-1503572639",
        "perseus-dummy_|-(none)_|-1.1_|-1.1_|-i586_|-(none)_|-1529936062",
    ]
    with patch.dict(zypper.__grains__, {"osarch": "x86_64"}), patch.dict(
        zypper.__salt__,
        {"cmd.run": MagicMock(return_value=os.linesep.join(rpm_out))},
    ), patch.dict(zypper.__salt__, {"pkg_resource.add_pkg": _add_data}), patch.dict(
        zypper.__salt__,
        {"pkg_resource.format_pkg_list": pkg_resource.format_pkg_list},
    ), patch.dict(
        zypper.__salt__, {"pkg_resource.stringify": MagicMock()}
    ), patch.object(
        zypper, "_list_pkgs_from_context"
    ) as list_pkgs_context_mock:
        pkgs = zypper.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()

        pkgs = zypper.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()


def test_normalize_name():
    """
    Test that package is normalized only when it should be
    """
    with patch.dict(zypper.__grains__, {"osarch": "x86_64"}):
        result = zypper.normalize_name("foo")
        assert result == "foo", result
        result = zypper.normalize_name("foo.x86_64")
        assert result == "foo", result
        result = zypper.normalize_name("foo.noarch")
        assert result == "foo", result

    with patch.dict(zypper.__grains__, {"osarch": "aarch64"}):
        result = zypper.normalize_name("foo")
        assert result == "foo", result
        result = zypper.normalize_name("foo.aarch64")
        assert result == "foo", result
        result = zypper.normalize_name("foo.noarch")
        assert result == "foo", result


def test_pkg_hold():
    """
    Tests holding packages with Zypper
    """

    # Test openSUSE 15.3
    list_locks_mock = {
        "bar": {"type": "package", "match_type": "glob", "case_sensitive": "on"},
        "minimal_base": {
            "type": "pattern",
            "match_type": "glob",
            "case_sensitive": "on",
        },
        "baz": {"type": "package", "match_type": "glob", "case_sensitive": "on"},
    }

    cmd = MagicMock(
        return_value={
            "pid": 1234,
            "retcode": 0,
            "stdout": "Specified lock has been successfully added.",
            "stderr": "",
        }
    )
    with patch.object(
        zypper, "list_locks", MagicMock(return_value=list_locks_mock)
    ), patch.dict(zypper.__salt__, {"cmd.run_all": cmd}):
        ret = zypper.hold("foo")
        assert ret["foo"]["changes"]["old"] == ""
        assert ret["foo"]["changes"]["new"] == "hold"
        assert ret["foo"]["comment"] == "Package foo is now being held."
        cmd.assert_called_once_with(
            ["zypper", "--non-interactive", "--no-refresh", "al", "foo"],
            env={},
            output_loglevel="trace",
            python_shell=False,
        )
        cmd.reset_mock()
        ret = zypper.hold(pkgs=["foo", "bar"])
        assert ret["foo"]["changes"]["old"] == ""
        assert ret["foo"]["changes"]["new"] == "hold"
        assert ret["foo"]["comment"] == "Package foo is now being held."
        assert ret["bar"]["changes"] == {}
        assert ret["bar"]["comment"] == "Package bar is already set to be held."
        cmd.assert_called_once_with(
            ["zypper", "--non-interactive", "--no-refresh", "al", "foo"],
            env={},
            output_loglevel="trace",
            python_shell=False,
        )


def test_pkg_unhold():
    """
    Tests unholding packages with Zypper
    """

    # Test openSUSE 15.3
    list_locks_mock = {
        "bar": {"type": "package", "match_type": "glob", "case_sensitive": "on"},
        "minimal_base": {
            "type": "pattern",
            "match_type": "glob",
            "case_sensitive": "on",
        },
        "baz": {"type": "package", "match_type": "glob", "case_sensitive": "on"},
    }

    cmd = MagicMock(
        return_value={
            "pid": 1234,
            "retcode": 0,
            "stdout": "1 lock has been successfully removed.",
            "stderr": "",
        }
    )
    with patch.object(
        zypper, "list_locks", MagicMock(return_value=list_locks_mock)
    ), patch.dict(zypper.__salt__, {"cmd.run_all": cmd}):
        ret = zypper.unhold("foo")
        assert ret["foo"]["comment"] == "Package foo was already unheld."
        cmd.assert_not_called()
        cmd.reset_mock()
        ret = zypper.unhold(pkgs=["foo", "bar"])
        assert ret["foo"]["changes"] == {}
        assert ret["foo"]["comment"] == "Package foo was already unheld."
        assert ret["bar"]["changes"]["old"] == "hold"
        assert ret["bar"]["changes"]["new"] == ""
        assert ret["bar"]["comment"] == "Package bar is no longer held."
        cmd.assert_called_once_with(
            ["zypper", "--non-interactive", "--no-refresh", "rl", "bar"],
            env={},
            output_loglevel="trace",
            python_shell=False,
        )


def test_pkg_list_holds():
    """
    Tests listing of calculated held packages with Zypper
    """

    # Test openSUSE 15.3
    list_locks_mock = {
        "bar": {"type": "package", "match_type": "glob", "case_sensitive": "on"},
        "minimal_base": {
            "type": "pattern",
            "match_type": "glob",
            "case_sensitive": "on",
        },
        "baz": {"type": "package", "match_type": "glob", "case_sensitive": "on"},
    }
    installed_pkgs = {
        "foo": [{"edition": "1.2.3-1.1"}],
        "bar": [{"edition": "2.3.4-2.1", "epoch": "2"}],
    }

    def zypper_search_mock(name, *_args, **_kwargs):
        if name in installed_pkgs:
            return {name: installed_pkgs.get(name)}

    with patch.object(
        zypper, "list_locks", MagicMock(return_value=list_locks_mock)
    ), patch.object(
        zypper, "search", MagicMock(side_effect=zypper_search_mock)
    ), patch.object(
        zypper, "info_installed", MagicMock(side_effect=zypper_search_mock)
    ):
        ret = zypper.list_holds()
        assert len(ret) == 1
        assert "bar-2:2.3.4-2.1.*" in ret


@pytest.mark.parametrize(
    "package,pre_version,post_version,fromrepo_param,name_param,pkgs_param,diff_attr_param",
    [
        ("vim", "1.1", "1.2", [], "", [], "all"),
        ("kernel-default", "1.1", "1.1,1.2", ["dummy", "dummy2"], "", [], None),
        ("vim", "1.1", "1.2", [], "vim", [], None),
    ],
)
def test_upgrade(
    package,
    pre_version,
    post_version,
    fromrepo_param,
    name_param,
    pkgs_param,
    diff_attr_param,
):
    with patch.object(zypper, "refresh_db", MagicMock(return_value=True)), patch(
        "salt.modules.zypperpkg.__zypper__.noraise.call"
    ) as zypper_mock, patch.object(
        zypper,
        "list_pkgs",
        MagicMock(side_effect=[{package: pre_version}, {package: post_version}]),
    ) as list_pkgs_mock:
        expected_call = ["update", "--auto-agree-with-licenses"]
        for repo in fromrepo_param:
            expected_call.extend(["--repo", repo])

        if pkgs_param:
            expected_call.extend(pkgs_param)
        elif name_param:
            expected_call.append(name_param)

        result = zypper.upgrade(
            name=name_param,
            pkgs=pkgs_param,
            fromrepo=fromrepo_param,
            diff_attr=diff_attr_param,
        )
        zypper_mock.assert_any_call(*expected_call)
        assert result == {package: {"old": pre_version, "new": post_version}}
        list_pkgs_mock.assert_any_call(root=None, attr=diff_attr_param)


@pytest.mark.parametrize(
    "package,pre_version,post_version,fromrepo_param",
    [
        ("vim", "1.1", "1.2", []),
        ("emacs", "1.1", "1.2", ["Dummy", "Dummy2"]),
    ],
)
def test_dist_upgrade(package, pre_version, post_version, fromrepo_param):
    with patch.object(zypper, "refresh_db", MagicMock(return_value=True)), patch(
        "salt.modules.zypperpkg.__zypper__.noraise.call"
    ) as zypper_mock, patch.object(
        zypper,
        "list_pkgs",
        MagicMock(side_effect=[{package: pre_version}, {package: post_version}]),
    ):
        expected_call = ["dist-upgrade", "--auto-agree-with-licenses"]

        for repo in fromrepo_param:
            expected_call.extend(["--from", repo])

        result = zypper.upgrade(dist_upgrade=True, fromrepo=fromrepo_param)
        zypper_mock.assert_any_call(*expected_call)
        assert result == {package: {"old": pre_version, "new": post_version}}


@pytest.mark.parametrize(
    "package,pre_version,post_version,fromrepo_param",
    [
        ("vim", "1.1", "1.1", []),
        ("emacs", "1.1", "1.1", ["Dummy", "Dummy2"]),
    ],
)
def test_dist_upgrade_dry_run(package, pre_version, post_version, fromrepo_param):
    with patch.object(zypper, "refresh_db", MagicMock(return_value=True)), patch(
        "salt.modules.zypperpkg.__zypper__.noraise.call"
    ) as zypper_mock, patch.object(
        zypper,
        "list_pkgs",
        MagicMock(side_effect=[{package: pre_version}, {package: post_version}]),
    ):
        expected_call = ["dist-upgrade", "--auto-agree-with-licenses", "--dry-run"]

        for repo in fromrepo_param:
            expected_call.extend(["--from", repo])

        zypper.upgrade(dist_upgrade=True, dryrun=True, fromrepo=fromrepo_param)
        zypper_mock.assert_any_call(*expected_call)
        # dryrun=True causes two calls, one with a trailing --debug-solver flag
        expected_call.append("--debug-solver")
        zypper_mock.assert_any_call(*expected_call)


def test_dist_upgrade_failure():
    zypper_output = textwrap.dedent(
        """\
        Loading repository data...
        Reading installed packages...
        Computing distribution upgrade...
        Use 'zypper repos' to get the list of defined repositories.
        Repository 'DUMMY' not found by its alias, number, or URI.
        """
    )
    call_spy = MagicMock()
    zypper_mock = MagicMock()
    zypper_mock.stdout = zypper_output
    zypper_mock.stderr = ""
    zypper_mock.exit_code = 3
    zypper_mock.noraise.call = call_spy
    with patch.object(zypper, "refresh_db", MagicMock(return_value=True)), patch(
        "salt.modules.zypperpkg.__zypper__", zypper_mock
    ), patch.object(
        zypper, "list_pkgs", MagicMock(side_effect=[{"vim": 1.1}, {"vim": 1.1}])
    ):
        expected_call = [
            "dist-upgrade",
            "--auto-agree-with-licenses",
            "--from",
            "Dummy",
        ]

        with pytest.raises(CommandExecutionError) as exc:
            zypper.upgrade(dist_upgrade=True, fromrepo=["Dummy"])
            call_spy.assert_called_with(*expected_call)

            assert exc.exception.info["changes"] == {}
            assert exc.exception.info["result"]["stdout"] == zypper_output



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.states.glusterfs
"""

import pytest

import salt.states.glusterfs as glusterfs
import salt.utils.cloud
import salt.utils.network
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {glusterfs: {}}


def test_peered():
    """
    Test to verify if node is peered.
    """
    name = "server1"

    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    mock_ip = MagicMock(return_value=["1.2.3.4", "1.2.3.5"])
    mock_ip6 = MagicMock(return_value=["2001:db8::1"])
    mock_host_ips = MagicMock(return_value=["1.2.3.5"])
    mock_peer = MagicMock(return_value=True)
    mock_status = MagicMock(return_value={"uuid1": {"hostnames": [name]}})

    with patch.dict(
        glusterfs.__salt__,
        {"glusterfs.peer_status": mock_status, "glusterfs.peer": mock_peer},
    ):
        with patch.object(salt.utils.network, "ip_addrs", mock_ip), patch.object(
            salt.utils.network, "ip_addrs6", mock_ip6
        ), patch.object(salt.utils.network, "host_to_ips", mock_host_ips):
            comt = "Peering with localhost is not needed"
            ret.update({"comment": comt})
            assert glusterfs.peered(name) == ret

            mock_host_ips.return_value = ["127.0.1.1"]
            comt = "Peering with localhost is not needed"
            ret.update({"comment": comt})
            assert glusterfs.peered(name) == ret

            mock_host_ips.return_value = ["2001:db8::1"]
            assert glusterfs.peered(name) == ret

            mock_host_ips.return_value = ["1.2.3.42"]
            comt = f"Host {name} already peered"
            ret.update({"comment": comt})
            assert glusterfs.peered(name) == ret

            with patch.dict(glusterfs.__opts__, {"test": False}):
                old = {"uuid1": {"hostnames": ["other1"]}}
                new = {
                    "uuid1": {"hostnames": ["other1"]},
                    "uuid2": {"hostnames": ["someAlias", name]},
                }
                mock_status.side_effect = [old, new]
                comt = f"Host {name} successfully peered"
                ret.update({"comment": comt, "changes": {"old": old, "new": new}})
                assert glusterfs.peered(name) == ret
                mock_status.side_effect = None

                mock_status.return_value = {"uuid1": {"hostnames": ["other"]}}
                mock_peer.return_value = False

                ret.update({"result": False})

                comt = "Failed to peer with {}, please check logs for errors".format(
                    name
                )
                ret.update({"comment": comt, "changes": {}})
                assert glusterfs.peered(name) == ret

                comt = "Invalid characters in peer name."
                ret.update({"comment": comt, "name": ":/"})
                assert glusterfs.peered(":/") == ret
                ret.update({"name": name})

            with patch.dict(glusterfs.__opts__, {"test": True}):
                comt = f"Peer {name} will be added."
                ret.update({"comment": comt, "result": None})
                assert glusterfs.peered(name) == ret


def test_volume_present():
    """
    Test to ensure that a volume exists
    """
    name = "salt"
    bricks = ["host1:/brick1"]
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    started_info = {name: {"status": "1"}}
    stopped_info = {name: {"status": "0"}}

    mock_info = MagicMock()
    mock_list = MagicMock()
    mock_create = MagicMock()
    mock_start = MagicMock(return_value=True)

    with patch.dict(
        glusterfs.__salt__,
        {
            "glusterfs.info": mock_info,
            "glusterfs.list_volumes": mock_list,
            "glusterfs.create_volume": mock_create,
            "glusterfs.start_volume": mock_start,
        },
    ):
        with patch.dict(glusterfs.__opts__, {"test": False}):
            mock_list.return_value = [name]
            mock_info.return_value = started_info
            comt = f"Volume {name} already exists and is started"
            ret.update({"comment": comt})
            assert glusterfs.volume_present(name, bricks, start=True) == ret

            mock_info.return_value = stopped_info
            comt = f"Volume {name} already exists and is now started"
            ret.update(
                {"comment": comt, "changes": {"old": "stopped", "new": "started"}}
            )
            assert glusterfs.volume_present(name, bricks, start=True) == ret

            comt = f"Volume {name} already exists"
            ret.update({"comment": comt, "changes": {}})
            assert glusterfs.volume_present(name, bricks, start=False) == ret
        with patch.dict(glusterfs.__opts__, {"test": True}):
            comt = f"Volume {name} already exists"
            ret.update({"comment": comt, "result": None})
            assert glusterfs.volume_present(name, bricks, start=False) == ret

            comt = f"Volume {name} already exists and will be started"
            ret.update({"comment": comt, "result": None})
            assert glusterfs.volume_present(name, bricks, start=True) == ret

            mock_list.return_value = []
            comt = f"Volume {name} will be created"
            ret.update({"comment": comt, "result": None})
            assert glusterfs.volume_present(name, bricks, start=False) == ret

            comt = f"Volume {name} will be created and started"
            ret.update({"comment": comt, "result": None})
            assert glusterfs.volume_present(name, bricks, start=True) == ret

        with patch.dict(glusterfs.__opts__, {"test": False}):
            mock_list.side_effect = [[], [name]]
            comt = f"Volume {name} is created"
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "changes": {"old": [], "new": [name]},
                }
            )
            assert glusterfs.volume_present(name, bricks, start=False) == ret

            mock_list.side_effect = [[], [name]]
            comt = f"Volume {name} is created and is now started"
            ret.update({"comment": comt, "result": True})
            assert glusterfs.volume_present(name, bricks, start=True) == ret

            mock_list.side_effect = None
            mock_list.return_value = []
            mock_create.return_value = False
            comt = f"Creation of volume {name} failed"
            ret.update({"comment": comt, "result": False, "changes": {}})
            assert glusterfs.volume_present(name, bricks) == ret

        with patch.object(salt.utils.cloud, "check_name", MagicMock(return_value=True)):
            comt = "Invalid characters in volume name."
            ret.update({"comment": comt, "result": False})
            assert glusterfs.volume_present(name, bricks) == ret


def test_started():
    """
    Test to check if volume has been started
    """
    name = "salt"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    started_info = {name: {"status": "1"}}
    stopped_info = {name: {"status": "0"}}
    mock_info = MagicMock(return_value={})
    mock_start = MagicMock(return_value=True)

    with patch.dict(
        glusterfs.__salt__,
        {"glusterfs.info": mock_info, "glusterfs.start_volume": mock_start},
    ):
        comt = f"Volume {name} does not exist"
        ret.update({"comment": comt})
        assert glusterfs.started(name) == ret

        mock_info.return_value = started_info
        comt = f"Volume {name} is already started"
        ret.update({"comment": comt, "result": True})
        assert glusterfs.started(name) == ret

        with patch.dict(glusterfs.__opts__, {"test": True}):
            mock_info.return_value = stopped_info
            comt = f"Volume {name} will be started"
            ret.update({"comment": comt, "result": None})
            assert glusterfs.started(name) == ret

        with patch.dict(glusterfs.__opts__, {"test": False}):
            comt = f"Volume {name} is started"
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "change": {"new": "started", "old": "stopped"},
                }
            )
            assert glusterfs.started(name) == ret


def test_add_volume_bricks():
    """
    Test to add brick(s) to an existing volume
    """
    name = "salt"
    bricks = ["host1:/drive1"]
    old_bricks = ["host1:/drive2"]

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    stopped_volinfo = {"salt": {"status": "0"}}
    volinfo = {"salt": {"status": "1", "bricks": {"brick1": {"path": old_bricks[0]}}}}
    new_volinfo = {
        "salt": {
            "status": "1",
            "bricks": {
                "brick1": {"path": old_bricks[0]},
                "brick2": {"path": bricks[0]},
            },
        }
    }

    mock_info = MagicMock(return_value={})
    mock_add = MagicMock(side_effect=[False, True])

    with patch.dict(
        glusterfs.__salt__,
        {"glusterfs.info": mock_info, "glusterfs.add_volume_bricks": mock_add},
    ):
        ret.update({"comment": "Volume salt does not exist"})
        assert glusterfs.add_volume_bricks(name, bricks) == ret

        mock_info.return_value = stopped_volinfo
        ret.update({"comment": "Volume salt is not started"})
        assert glusterfs.add_volume_bricks(name, bricks) == ret

        mock_info.return_value = volinfo
        ret.update({"comment": "Adding bricks to volume salt failed"})
        assert glusterfs.add_volume_bricks(name, bricks) == ret

        ret.update({"result": True})
        ret.update({"comment": "Bricks already added in volume salt"})
        assert glusterfs.add_volume_bricks(name, old_bricks) == ret

        mock_info.side_effect = [volinfo, new_volinfo]
        ret.update(
            {
                "comment": "Bricks successfully added to volume salt",
                "changes": {"new": bricks + old_bricks, "old": old_bricks},
            }
        )
        # Let's sort ourselves because the test under python 3 sometimes fails
        # just because of the new changes list order
        result = glusterfs.add_volume_bricks(name, bricks)
        ret["changes"]["new"] = sorted(ret["changes"]["new"])
        result["changes"]["new"] = sorted(result["changes"]["new"])
        assert result == ret


def test_op_version():
    """
    Test setting the Glusterfs op-version
    """
    name = "salt"
    current = 30707
    new = 31200

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    mock_get_version = MagicMock(return_value={})
    mock_set_version = MagicMock(return_value={})

    with patch.dict(
        glusterfs.__salt__,
        {
            "glusterfs.get_op_version": mock_get_version,
            "glusterfs.set_op_version": mock_set_version,
        },
    ):
        mock_get_version.return_value = [False, "some error message"]
        ret.update({"result": False})
        ret.update({"comment": "some error message"})
        assert glusterfs.op_version(name, current) == ret

        mock_get_version.return_value = current
        ret.update({"result": True})
        ret.update(
            {
                "comment": (
                    "Glusterfs cluster.op-version for {} already set to {}".format(
                        name, current
                    )
                )
            }
        )
        assert glusterfs.op_version(name, current) == ret

        with patch.dict(glusterfs.__opts__, {"test": True}):
            mock_set_version.return_value = [False, "Failed to set version"]
            ret.update({"result": None})
            ret.update(
                {
                    "comment": (
                        "An attempt would be made to set the cluster.op-version for"
                        " {} to {}.".format(name, new)
                    )
                }
            )
            assert glusterfs.op_version(name, new) == ret

        with patch.dict(glusterfs.__opts__, {"test": False}):
            mock_set_version.return_value = [False, "Failed to set version"]
            ret.update({"result": False})
            ret.update({"comment": "Failed to set version"})
            assert glusterfs.op_version(name, new) == ret

            mock_set_version.return_value = "some success message"
            ret.update({"comment": "some success message"})
            ret.update({"changes": {"old": current, "new": new}})
            ret.update({"result": True})
            assert glusterfs.op_version(name, new) == ret


def test_max_op_version():
    """
    Test setting the Glusterfs to its self reported max-op-version
    """
    name = "salt"
    current = 30707
    new = 31200

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    mock_get_version = MagicMock(return_value={})
    mock_get_max_op_version = MagicMock(return_value={})
    mock_set_version = MagicMock(return_value={})

    with patch.dict(
        glusterfs.__salt__,
        {
            "glusterfs.get_op_version": mock_get_version,
            "glusterfs.set_op_version": mock_set_version,
            "glusterfs.get_max_op_version": mock_get_max_op_version,
        },
    ):
        mock_get_version.return_value = [False, "some error message"]
        ret.update({"result": False})
        ret.update({"comment": "some error message"})
        assert glusterfs.max_op_version(name) == ret

        mock_get_version.return_value = current
        mock_get_max_op_version.return_value = [False, "some error message"]
        ret.update({"result": False})
        ret.update({"comment": "some error message"})
        assert glusterfs.max_op_version(name) == ret

        mock_get_version.return_value = current
        mock_get_max_op_version.return_value = current
        ret.update({"result": True})
        ret.update(
            {
                "comment": (
                    "The cluster.op-version is already set to the"
                    " cluster.max-op-version of {}".format(current)
                )
            }
        )
        assert glusterfs.max_op_version(name) == ret

        with patch.dict(glusterfs.__opts__, {"test": True}):
            mock_get_max_op_version.return_value = new
            ret.update({"result": None})
            ret.update(
                {
                    "comment": (
                        "An attempt would be made to set the cluster.op-version"
                        " to {}.".format(new)
                    )
                }
            )
            assert glusterfs.max_op_version(name) == ret

        with patch.dict(glusterfs.__opts__, {"test": False}):
            mock_set_version.return_value = [False, "Failed to set version"]
            ret.update({"result": False})
            ret.update({"comment": "Failed to set version"})
            assert glusterfs.max_op_version(name) == ret

            mock_set_version.return_value = "some success message"
            ret.update({"comment": "some success message"})
            ret.update({"changes": {"old": current, "new": new}})
            ret.update({"result": True})
            assert glusterfs.max_op_version(name) == ret
